@@ -300,6 +300,7 @@ KNOWN_MIME_TYPES = {
     'application/x-dos-batch': 'bat batch cmd dos',
     'application/x-sh':        'sh',
     'application/x-csh':       'csh',
+    'application/x-genshi':    'genshi',
     'application/x-troff':     'nroff roff troff',
     'application/x-yaml':      'yml yaml',
 
@@ -319,6 +320,7 @@ KNOWN_MIME_TYPES = {
     'text/xml':                'xml',
 
     # see also TEXT_X_TYPES below
+    'text/x-apacheconf':       'apache',
     'text/x-csrc':             'c xs',
     'text/x-chdr':             'h',
     'text/x-c++src':           'cc CC cpp C c++ C++',
@@ -330,6 +332,7 @@ KNOWN_MIME_TYPES = {
     'text/x-fortran':          'f',
     'text/x-haskell':          'hs',
     'text/x-ini':              'ini cfg',
+    'text/x-nginx-conf':       'nginx',
     'text/x-objc':             'm mm',
     'text/x-ocaml':            'ml mli',
     'text/x-makefile':         'make mk Makefile GNUMakefile',
@@ -337,6 +340,7 @@ KNOWN_MIME_TYPES = {
     'text/x-perl':             'pl pm PL',
     'text/x-php':              'php3 php4',
     'text/x-python':           'py',
+    'text/x-python-doctest':   'pycon',
     'text/x-pyrex':            'pyx',
     'text/x-ruby':             'rb',
     'text/x-scheme':           'scm',
@@ -653,7 +657,7 @@ class Mimeview(Component):
 
     treat_as_binary = ListOption('mimeviewer', 'treat_as_binary',
         'application/octet-stream, application/pdf, application/postscript, '
-        'application/msword,application/rtf,',
+        'application/msword, application/rtf',
         doc="""Comma-separated list of MIME types that should be treated as
         binary data. (''since 0.11.5'')""")
 
@@ -822,8 +826,8 @@ class Mimeview(Component):
                     return tag.div(class_='code')(tag.pre(result)).generate()
 
             except Exception, e:
-                self.log.warning('HTML preview using %s failed: %s',
-                                 renderer.__class__.__name__,
+                self.log.warning('HTML preview using %s with %r failed: %s',
+                                 renderer.__class__.__name__, context,
                                  exception_to_unicode(e, traceback=True))
                 if context.req and not context.get_hint('disable_warnings'):
                     from trac.web.chrome import add_warning
