@@ -18,16 +18,20 @@
 
 from __future__ import with_statement
 
+from contextlib import contextmanager
 from hashlib import md5
 
 from genshi.template.text import NewTextTemplate
 
+from trac.attachment import IAttachmentChangeListener
 from trac.core import *
 from trac.config import *
 from trac.notification import NotifyEmail
 from trac.ticket.api import TicketSystem
-from trac.util.datefmt import to_utimestamp
-from trac.util.text import obfuscate_email_address, text_width, wrap
+from trac.ticket.model import Ticket
+from trac.util.datefmt import datetime_now, to_utimestamp, utc
+from trac.util.text import exception_to_unicode, obfuscate_email_address, \
+                           shorten_line, text_width, wrap
 from trac.util.translation import deactivate, reactivate
 
 
@@ -57,7 +61,7 @@ class TicketNotificationSystem(Component):
         ''(since 0.11)''""")
 
     batch_subject_template = Option('notification', 'batch_subject_template',
-                                     '$prefix Batch modify: $tickets_descr',
+                                    '$prefix Batch modify: $tickets_descr',
         """Like ticket_subject_template but for batch modifications.
 
         By default, the template is `$prefix Batch modify: $tickets_descr`.
@@ -73,117 +77,117 @@ class TicketNotificationSystem(Component):
         US-ASCII characters.  This is expected by CJK users. ''(since
         0.12.2)''""")
 
-def get_ticket_notification_recipients(env, config, tktid, prev_cc):
-    notify_reporter = config.getbool('notification', 'always_notify_reporter')
-    notify_owner = config.getbool('notification', 'always_notify_owner')
-    notify_updater = config.getbool('notification', 'always_notify_updater')
 
-    ccrecipients = prev_cc
-    torecipients = []
-    with env.db_query as db:
-        # Harvest email addresses from the cc, reporter, and owner fields
-        for row in db("SELECT cc, reporter, owner FROM ticket WHERE id=%s",
-                      (tktid,)):
-            if row[0]:
-                ccrecipients += row[0].replace(',', ' ').split()
-            reporter = row[1]
-            owner = row[2]
-            if notify_reporter:
-                torecipients.append(row[1])
-            if notify_owner:
-                torecipients.append(row[2])
-            break
+def get_ticket_notification_recipients(env, config, tktid, prev_cc=None,
+                                       modtime=None):
+    """Returns notifications recipients.
 
-        # Harvest email addresses from the author field of ticket_change(s)
-        if notify_updater:
-            for author, ticket in db("""
-                    SELECT DISTINCT author, ticket FROM ticket_change
-                    WHERE ticket=%s
-                    """, (tktid,)):
-                torecipients.append(author)
+    :since 1.0.2: the `config` parameter is no longer used.
+    :since 1.0.2: the `prev_cc` parameter is deprecated.
+    """
+    section = env.config['notification']
+    always_notify_reporter = section.getbool('always_notify_reporter')
+    always_notify_owner = section.getbool('always_notify_owner')
+    always_notify_updater = section.getbool('always_notify_updater')
 
-        # Suppress the updater from the recipients
-        updater = None
-        for updater, in db("""
-                SELECT author FROM ticket_change WHERE ticket=%s
-                ORDER BY time DESC LIMIT 1
-                """, (tktid,)):
-            break
-        else:
-            for updater, in db("SELECT reporter FROM ticket WHERE id=%s",
-                               (tktid,)):
-                break
+    cc_recipients = set(prev_cc or [])
+    to_recipients = set()
+    tkt = Ticket(env, tktid)
 
-        if not notify_updater:
-            filter_out = True
-            if notify_reporter and (updater == reporter):
-                filter_out = False
-            if notify_owner and (updater == owner):
-                filter_out = False
-            if filter_out:
-                torecipients = [r for r in torecipients
-                                if r and r != updater]
-        elif updater:
-            torecipients.append(updater)
+    # CC field is stored as comma-separated string. Parse to list.
+    to_list = lambda cc: cc.replace(',', ' ').split()
 
-    return (torecipients, ccrecipients, reporter, owner)
+    # Backward compatibility
+    if not modtime:
+        modtime = tkt['changetime']
+
+    # Harvest email addresses from the author field of ticket_change(s)
+    if always_notify_updater:
+        for author, ticket in env.db_query("""
+                SELECT DISTINCT author, ticket FROM ticket_change
+                WHERE ticket=%s
+                """, (tktid, )):
+            to_recipients.add(author)
+
+    # Harvest previous owner and cc list
+    author = None
+    for changelog in tkt.get_changelog(modtime):
+        author, field, old = changelog[1:4]
+        if field == 'owner' and always_notify_owner:
+            to_recipients.add(old)
+        elif field == 'cc':
+            cc_recipients.update(to_list(old))
+
+    # Harvest email addresses from the cc, reporter, and owner fields
+    updater = author or tkt['reporter']
+    if tkt['cc']:
+        cc_recipients.update(to_list(tkt['cc']))
+    if always_notify_reporter:
+        to_recipients.add(tkt['reporter'])
+    if always_notify_owner:
+        to_recipients.add(tkt['owner'])
+    if always_notify_updater and updater:
+        to_recipients.add(updater)
+
+    # Suppress the updater from the recipients if necessary
+    if not always_notify_updater:
+        filter_out = True
+        if always_notify_reporter and updater == tkt['reporter']:
+            filter_out = False
+        if always_notify_owner and updater == tkt['owner']:
+            filter_out = False
+        if filter_out:
+            to_recipients.discard(updater)
+
+    return list(to_recipients), list(cc_recipients), \
+           tkt['reporter'], tkt['owner']
 
 
 class TicketNotifyEmail(NotifyEmail):
     """Notification of ticket changes."""
 
     template_name = "ticket_notify_email.txt"
-    ticket = None
-    newticket = None
-    modtime = 0
     from_email = 'trac+ticket@localhost'
     COLS = 75
 
     def __init__(self, env):
-        NotifyEmail.__init__(self, env)
-        self.prev_cc = []
+        super(TicketNotifyEmail, self).__init__(env)
         ambiguous_char_width = env.config.get('notification',
                                               'ambiguous_char_width',
                                               'single')
         self.ambiwidth = 2 if ambiguous_char_width == 'double' else 1
+        self.ticket = None
+        self.modtime = 0
+        self.newticket = None
+        self.reporter = None
+        self.owner = None
 
     def notify(self, ticket, newticket=True, modtime=None):
         """Send ticket change notification e-mail (untranslated)"""
-        t = deactivate()
-        translated_fields = ticket.fields
-        try:
-            ticket.fields = TicketSystem(self.env).get_ticket_fields()
-            self._notify(ticket, newticket, modtime)
-        finally:
-            ticket.fields = translated_fields
-            reactivate(t)
-
-    def _notify(self, ticket, newticket=True, modtime=None):
         self.ticket = ticket
         self.modtime = modtime
         self.newticket = newticket
-
-        changes_body = ''
         self.reporter = ''
         self.owner = ''
-        changes_descr = ''
-        change_data = {}
         link = self.env.abs_href.ticket(ticket.id)
         summary = self.ticket['summary']
         author = None
 
+        changes_body = ''
+        changes_descr = ''
+        change_data = {}
         if not self.newticket and modtime:  # Ticket change
             from trac.ticket.web_ui import TicketModule
-            for change in TicketModule(self.env).grouped_changelog_entries(
-                                                ticket, when=modtime):
-                if not change['permanent']: # attachment with same time...
+            for change in TicketModule(self.env) \
+                          .grouped_changelog_entries(ticket, when=modtime):
+                if not change['permanent']:  # attachment with same time...
                     continue
                 author = change['author']
                 change_data.update({
                     'author': self.obfuscate_email(author),
                     'comment': wrap(change['comment'], self.COLS, ' ', ' ',
                                     '\n', self.ambiwidth)
-                    })
+                })
                 link += '#comment:%s' % str(change.get('cnum', ''))
                 for field, values in change['fields'].iteritems():
                     old = values['old']
@@ -194,7 +198,7 @@ class TicketNotifyEmail(NotifyEmail):
                                          self.ambiwidth)
                         old_descr = wrap(old, self.COLS, '> ', '> ', '\n',
                                          self.ambiwidth)
-                        old_descr = old_descr.replace(2 * '\n', '\n' + '>' + \
+                        old_descr = old_descr.replace(2 * '\n', '\n' + '>' +
                                                       '\n')
                         cdescr = '\n'
                         cdescr += 'Old description:' + 2 * '\n' + old_descr + \
@@ -205,7 +209,7 @@ class TicketNotifyEmail(NotifyEmail):
                     elif field == 'summary':
                         summary = "%s (was: %s)" % (new, old)
                     elif field == 'cc':
-                        (addcc, delcc) = self.diff_cc(old, new)
+                        addcc, delcc = self.diff_cc(old, new)
                         chgcc = ''
                         if delcc:
                             chgcc += wrap(" * cc: %s (removed)" %
@@ -219,7 +223,6 @@ class TicketNotifyEmail(NotifyEmail):
                                           self.ambiwidth) + '\n'
                         if chgcc:
                             changes_body += chgcc
-                        self.prev_cc += self.parse_cc(old) if old else []
                     else:
                         if field in ['owner', 'reporter']:
                             old = self.obfuscate_email(old)
@@ -255,19 +258,61 @@ class TicketNotifyEmail(NotifyEmail):
         ticket_values['new'] = self.newticket
         ticket_values['link'] = link
 
-        subject = self.format_subj(summary)
-        if not self.newticket:
-            subject = 'Re: ' + subject
-        self.data.update({
-            'ticket_props': self.format_props(),
-            'ticket_body_hdr': self.format_hdr(),
-            'subject': subject,
-            'ticket': ticket_values,
-            'changes_body': changes_body,
-            'changes_descr': changes_descr,
-            'change': change_data
+        subject = self.format_subj(summary, newticket)
+
+        with _translation_deactivated(ticket):
+            self.data.update({
+                'ticket_props': self.format_props(),
+                'ticket_body_hdr': self.format_hdr(),
+                'subject': subject,
+                'ticket': ticket_values,
+                'changes_body': changes_body,
+                'changes_descr': changes_descr,
+                'change': change_data
             })
-        NotifyEmail.notify(self, ticket.id, subject, author)
+            super(TicketNotifyEmail, self).notify(ticket.id, subject, author)
+
+    def notify_attachment(self, ticket, attachment, added=True):
+        """Send ticket attachment notification (untranslated)"""
+        self.ticket = ticket
+        self.modtime = attachment.date or datetime_now(utc)
+        self.newticket = False
+        self.reporter = ''
+        self.owner = ''
+        link = self.env.abs_href.ticket(ticket.id)
+        summary = self.ticket['summary']
+        author = attachment.author
+
+        # Note: no translation yet
+        changes_body = wrap(" * Attachment \"%s\" %s."
+                            % (attachment.filename,
+                               "added" if added else "removed"),
+                            self.COLS, ' ', ' ', '\n',
+                            self.ambiwidth) + "\n"
+        if attachment.description:
+            changes_body += "\n" + wrap(attachment.description, self.COLS,
+                                        ' ', ' ', '\n', self.ambiwidth)
+
+        ticket_values = ticket.values.copy()
+        ticket_values['id'] = ticket.id
+        ticket_values['description'] = wrap(
+            ticket_values.get('description', ''), self.COLS,
+            initial_indent=' ', subsequent_indent=' ', linesep='\n',
+            ambiwidth=self.ambiwidth)
+        ticket_values['new'] = self.newticket
+        ticket_values['link'] = link
+        subject = self.format_subj(summary, False)
+        with _translation_deactivated(ticket):
+            self.data.update({
+                'ticket_props': self.format_props(),
+                'ticket_body_hdr': self.format_hdr(),
+                'subject': subject,
+                'ticket': ticket_values,
+                'changes_body': changes_body,
+                'changes_descr': '',
+                'change': {'author': self.obfuscate_email(author)},
+            })
+            super(TicketNotifyEmail, self).notify(ticket.id, subject, author)
 
     def format_props(self):
         tkt = self.ticket
@@ -279,7 +324,7 @@ class TicketNotifyEmail(NotifyEmail):
             if f['type'] == 'textarea':
                 continue
             fname = f['name']
-            if not fname in tkt.values:
+            if fname not in tkt.values:
                 continue
             fval = tkt[fname] or ''
             if fval.find('\n') != -1:
@@ -306,13 +351,13 @@ class TicketNotifyEmail(NotifyEmail):
                 width_l = self.COLS - width_r - 1
         sep = width_l * '-' + '+' + width_r * '-'
         txt = sep + '\n'
-        cell_tmp = [u'', u'']
+        vals_lr = ([], [])
         big = []
         i = 0
         width_lr = [width_l, width_r]
         for f in [f for f in fields if f['name'] != 'description']:
             fname = f['name']
-            if not tkt.values.has_key(fname):
+            if fname not in tkt.values:
                 continue
             fval = tkt[fname] or ''
             if fname in ['owner', 'reporter']:
@@ -324,15 +369,36 @@ class TicketNotifyEmail(NotifyEmail):
                 # __str__ method won't be called.
                 str_tmp = u'%s:  %s' % (f['label'], unicode(fval))
                 idx = i % 2
-                cell_tmp[idx] += wrap(str_tmp, width_lr[idx] - 2 + 2 * idx,
-                                      (width[2 * idx]
-                                       - self.get_text_width(f['label'])
-                                       + 2 * idx) * ' ',
-                                      2 * ' ', '\n', self.ambiwidth)
-                cell_tmp[idx] += '\n'
+                initial_indent = ' ' * (width[2 * idx] -
+                                        self.get_text_width(f['label']) +
+                                        2 * idx)
+                wrapped = wrap(str_tmp, width_lr[idx] - 2 + 2 * idx,
+                               initial_indent, '  ', '\n', self.ambiwidth)
+                vals_lr[idx].append(wrapped.splitlines())
                 i += 1
-        cell_l = cell_tmp[0].splitlines()
-        cell_r = cell_tmp[1].splitlines()
+        if len(vals_lr[0]) > len(vals_lr[1]):
+            vals_lr[1].append([])
+
+        cell_l = []
+        cell_r = []
+        for i in xrange(len(vals_lr[0])):
+            vals_l = vals_lr[0][i]
+            vals_r = vals_lr[1][i]
+            vals_diff = len(vals_l) - len(vals_r)
+            diff = len(cell_l) - len(cell_r)
+            if diff > 0:
+                # add padding to right side if needed
+                if vals_diff < 0:
+                    diff += vals_diff
+                cell_r.extend([''] * max(diff, 0))
+            elif diff < 0:
+                # add padding to left side if needed
+                if vals_diff > 0:
+                    diff += vals_diff
+                cell_l.extend([''] * max(-diff, 0))
+            cell_l.extend(vals_l)
+            cell_r.extend(vals_r)
+
         for i in range(max(len(cell_l), len(cell_r))):
             if i >= len(cell_l):
                 cell_l.append(width_l * ' ')
@@ -354,19 +420,19 @@ class TicketNotifyEmail(NotifyEmail):
     def diff_cc(self, old, new):
         oldcc = NotifyEmail.addrsep_re.split(old)
         newcc = NotifyEmail.addrsep_re.split(new)
-        added = [self.obfuscate_email(x) \
-                                for x in newcc if x and x not in oldcc]
-        removed = [self.obfuscate_email(x) \
-                                for x in oldcc if x and x not in newcc]
-        return (added, removed)
+        added = [self.obfuscate_email(x)
+                 for x in newcc if x and x not in oldcc]
+        removed = [self.obfuscate_email(x)
+                   for x in oldcc if x and x not in newcc]
+        return added, removed
 
     def format_hdr(self):
         return '#%s: %s' % (self.ticket.id, wrap(self.ticket['summary'],
                                                  self.COLS, linesep='\n',
                                                  ambiwidth=self.ambiwidth))
 
-    def format_subj(self, summary):
-        template = self.config.get('notification','ticket_subject_template')
+    def format_subj(self, summary, newticket=True):
+        template = self.config.get('notification', 'ticket_subject_template')
         template = NewTextTemplate(template.encode('utf8'))
 
         prefix = self.config.get('notification', 'smtp_subject_prefix')
@@ -380,15 +446,18 @@ class TicketNotifyEmail(NotifyEmail):
             'env': self.env,
         }
 
-        return template.generate(**data).render('text', encoding=None).strip()
+        subj = template.generate(**data).render('text', encoding=None).strip()
+        if not newticket:
+            subj = "Re: " + subj
+        return subj
 
     def get_recipients(self, tktid):
-        (torecipients, ccrecipients, reporter, owner) = \
-            get_ticket_notification_recipients(self.env, self.config,
-                tktid, self.prev_cc)
+        to_recipients, cc_recipients, reporter, owner = \
+            get_ticket_notification_recipients(self.env, self.config, tktid,
+                                               modtime=self.modtime)
         self.reporter = reporter
         self.owner = owner
-        return (torecipients, ccrecipients)
+        return to_recipients, cc_recipients
 
     def get_message_id(self, rcpt, modtime=None):
         """Generate a predictable, but sufficiently unique message ID."""
@@ -402,15 +471,16 @@ class TicketNotifyEmail(NotifyEmail):
 
     def send(self, torcpts, ccrcpts):
         dest = self.reporter or 'anonymous'
-        hdrs = {}
-        hdrs['Message-ID'] = self.get_message_id(dest, self.modtime)
-        hdrs['X-Trac-Ticket-ID'] = str(self.ticket.id)
-        hdrs['X-Trac-Ticket-URL'] = self.data['ticket']['link']
+        hdrs = {
+            'Message-ID': self.get_message_id(dest, self.modtime),
+            'X-Trac-Ticket-ID': str(self.ticket.id),
+            'X-Trac-Ticket-URL': self.data['ticket']['link']
+        }
         if not self.newticket:
             msgid = self.get_message_id(dest)
             hdrs['In-Reply-To'] = msgid
             hdrs['References'] = msgid
-        NotifyEmail.send(self, torcpts, ccrcpts, hdrs)
+        super(TicketNotifyEmail, self).send(torcpts, ccrcpts, hdrs)
 
     def get_text_width(self, text):
         return text_width(text, ambiwidth=self.ambiwidth)
@@ -425,32 +495,66 @@ class TicketNotifyEmail(NotifyEmail):
         else:
             return obfuscate_email_address(text)
 
+
+class TicketAttachmentNotifier(Component):
+    """Sends notification on attachment change."""
+
+    implements(IAttachmentChangeListener)
+
+    # IAttachmentChangeListener methods
+
+    def attachment_added(self, attachment):
+        self._notify_attachment(attachment, True)
+
+    def attachment_deleted(self, attachment):
+        self._notify_attachment(attachment, False)
+
+    def attachment_reparented(self, attachment, old_parent_realm,
+                              old_parent_id):
+        pass
+
+    # Internal methods
+
+    def _notify_attachment(self, attachment, added):
+        resource = attachment.resource.parent
+        if resource.realm != 'ticket':
+            return
+        ticket = Ticket(self.env, resource.id)
+        tn = TicketNotifyEmail(self.env)
+        try:
+            tn.notify_attachment(ticket, attachment, added)
+        except Exception, e:
+            self.log.error("Failure sending notification when adding "
+                           "attachment %s to ticket #%s: %s",
+                           attachment.filename, ticket.id,
+                           exception_to_unicode(e))
+
+
 class BatchTicketNotifyEmail(NotifyEmail):
     """Notification of ticket batch modifications."""
 
     template_name = "batch_ticket_notify_email.txt"
 
     def __init__(self, env):
-        NotifyEmail.__init__(self, env)
+        super(BatchTicketNotifyEmail, self).__init__(env)
 
-    def notify(self, tickets, new_values, comment, action, author):
+    def notify(self, tickets, new_values, comment, action, author,
+               modtime=None):
         """Send batch ticket change notification e-mail (untranslated)"""
-        t = deactivate()
-        try:
-            self._notify(tickets, new_values, comment, action, author)
-        finally:
-            reactivate(t)
+        tickets = self._sort_tickets_by_priority(tickets)
+        with _translation_deactivated():
+            self._notify(tickets, new_values, comment, action, author, modtime)
 
-    def _notify(self, tickets, new_values, comment, action, author):
+    def _notify(self, tickets, new_values, comment, action, author, modtime):
         self.tickets = tickets
-        changes_body = ''
         self.reporter = ''
         self.owner = ''
-        changes_descr = '\n'.join(['%s to %s' % (prop, val)
-                                  for (prop, val) in new_values.iteritems()])
-        tickets_descr = ', '.join(['#%s' % t for t in tickets])
+        self.modtime = modtime
+        changes_descr = '\n'.join('%s to %s' % (prop, val)
+                                  for prop, val in new_values.iteritems())
+        tickets_descr = ', '.join('#%s' % t for t in tickets)
         subject = self.format_subj(tickets_descr)
-        link = self.env.abs_href.query(id=','.join([str(t) for t in tickets]))
+        link = self.env.abs_href.query(id=','.join(str(t) for t in tickets))
         self.data.update({
             'tickets_descr': tickets_descr,
             'changes_descr': changes_descr,
@@ -459,11 +563,25 @@ class BatchTicketNotifyEmail(NotifyEmail):
             'author': author,
             'subject': subject,
             'ticket_query_link': link,
-            })
-        NotifyEmail.notify(self, tickets, subject, author)
+        })
+        super(BatchTicketNotifyEmail, self).notify(tickets, subject, author)
+
+    def _sort_tickets_by_priority(self, tickets):
+        with self.env.db_query as db:
+            tickets = [int(id_) for id_ in tickets]
+            holders = ','.join(['%s'] * len(tickets))
+            rows = db("""
+                SELECT id FROM ticket AS t
+                LEFT OUTER JOIN enum p
+                    ON p.type='priority' AND p.name=t.priority
+                WHERE t.id IN (%s)
+                ORDER BY COALESCE(p.value,'')='', %s, t.id
+                """ % (holders, db.cast('p.value', 'int')),
+                tickets)
+            return [row[0] for row in rows]
 
     def format_subj(self, tickets_descr):
-        template = self.config.get('notification','batch_subject_template')
+        template = self.config.get('notification', 'batch_subject_template')
         template = NewTextTemplate(template.encode('utf8'))
 
         prefix = self.config.get('notification', 'smtp_subject_prefix')
@@ -475,16 +593,44 @@ class BatchTicketNotifyEmail(NotifyEmail):
             'tickets_descr': tickets_descr,
             'env': self.env,
         }
-
-        return template.generate(**data).render('text', encoding=None).strip()
+        subj = template.generate(**data).render('text', encoding=None).strip()
+        return shorten_line(subj)
 
     def get_recipients(self, tktids):
-        alltorecipients = []
-        allccrecipients = []
+        all_to_recipients = set()
+        all_cc_recipients = set()
         for t in tktids:
-            (torecipients, ccrecipients, reporter, owner) = \
-                get_ticket_notification_recipients(self.env, self.config,
-                    t, [])
-            alltorecipients.extend(torecipients)
-            allccrecipients.extend(ccrecipients)
-        return (list(set(alltorecipients)), list(set(allccrecipients)))
+            to_recipients, cc_recipients, reporter, owner = \
+                get_ticket_notification_recipients(self.env, self.config, t,
+                                                   modtime=self.modtime)
+            all_to_recipients.update(to_recipients)
+            all_cc_recipients.update(cc_recipients)
+        return list(all_to_recipients), list(all_cc_recipients)
+
+    def get_message_id(self, modtime=None):
+        s = '%s.%s.%d' % (self.env.project_url.encode('utf-8'),
+                          ','.join(map(str, self.tickets)),
+                          to_utimestamp(modtime))
+        dig = md5(s).hexdigest()
+        host = self.from_email[self.from_email.find('@') + 1:]
+        msgid = '<%03d.%s@%s>' % (len(s), dig, host)
+        return msgid
+
+    def send(self, torcpts, ccrcpts):
+        hdrs = {'Message-ID': self.get_message_id(self.modtime)}
+        super(BatchTicketNotifyEmail, self).send(torcpts, ccrcpts, hdrs)
+
+
+@contextmanager
+def _translation_deactivated(ticket=None):
+    t = deactivate()
+    if ticket is not None:
+        ts = TicketSystem(ticket.env)
+        translated_fields = ticket.fields
+        ticket.fields = ts.get_ticket_fields()
+    try:
+        yield
+    finally:
+        if ticket is not None:
+            ticket.fields = translated_fields
+        reactivate(t)
