@@ -34,8 +34,9 @@ from trac.resource import Resource
 from trac.ticket.api import TicketSystem
 from trac.ticket.model import Milestone, group_milestones
 from trac.util import Ranges, as_bool
-from trac.util.datefmt import format_datetime, from_utimestamp, parse_date, \
-                              to_timestamp, to_utimestamp, utc, user_time
+from trac.util.datefmt import datetime_now, format_datetime, from_utimestamp, \
+                              parse_date, to_timestamp, to_utimestamp, utc, \
+                              user_time
 from trac.util.presentation import Paginator
 from trac.util.text import empty, shorten_line, quote_query_string
 from trac.util.translation import _, tag_, cleandoc_, ngettext
@@ -46,6 +47,7 @@ from trac.web.chrome import (INavigationContributor, Chrome,
                              add_script_data, add_stylesheet, add_warning,
                              web_context)
 from trac.wiki.api import IWikiSyntaxProvider
+from trac.wiki.formatter import system_message
 from trac.wiki.macros import WikiMacroBase # TODO: should be moved in .api
 
 
@@ -335,10 +337,7 @@ class Query(object):
                     elif name in self.time_fields:
                         val = from_utimestamp(val)
                     elif field and field['type'] == 'checkbox':
-                        try:
-                            val = bool(int(val))
-                        except (TypeError, ValueError):
-                            val = False
+                        val = as_bool(val)
                     elif val is None:
                         val = ''
                     result[name] = val
@@ -430,7 +429,6 @@ class Query(object):
         self.get_columns()
         db = self.env.get_read_db()
 
-        enum_columns = ('resolution', 'priority', 'severity')
         # Build the list of actual columns to query
         cols = []
         def add_cols(*args):
@@ -445,18 +443,26 @@ class Query(object):
         add_cols('status', 'priority', 'time', 'changetime', self.order)
         cols.extend([c for c in self.constraint_cols if not c in cols])
 
-        custom_fields = [f['name'] for f in self.fields if f.get('custom')]
-        list_fields = [f['name'] for f in self.fields
-                                 if f['type'] == 'text' and
-                                    f.get('format') == 'list']
-        # 32 is max of joins in SQLite 32-bit, 3 is for order, group and
-        # "priority" columns
-        use_joins = len(set(cols) & set(custom_fields)) <= 32 - 3
+        custom_fields = set(f['name'] for f in self.fields if f.get('custom'))
+        list_fields = set(f['name'] for f in self.fields
+                                    if f['type'] == 'text' and
+                                       f.get('format') == 'list')
+        enum_columns = [col for col in ('resolution', 'priority', 'severity',
+                                        'type')
+                            if col not in custom_fields and
+                               col in ('priority', self.order, self.group)]
+        joined_columns = [col for col in ('milestone', 'version')
+                              if col not in custom_fields and
+                                 col in (self.order, self.group)]
+        # 31 is max of joins in SQLite 32-bit
+        use_joins = (len(set(cols) & custom_fields) +
+                     len(enum_columns) + len(joined_columns)) <= 31
 
         sql = []
-        sql.append("SELECT " + ",".join(['t.%s AS %s' % (c, c) for c in cols
-                                         if c not in custom_fields]))
-        sql.append(",priority.value AS priority_value")
+        sql.append("SELECT " + ",".join('t.%s AS %s' % (c, c) for c in cols
+                                        if c not in custom_fields))
+        if 'priority' in enum_columns:
+            sql.append(",priority.value AS priority_value")
 
         if use_joins:
             # Use LEFT OUTER JOIN for ticket_custom table
@@ -481,17 +487,15 @@ class Query(object):
             sql.append("\n  FROM ticket AS t) AS t")
 
         # Join with the enum table for proper sorting
-        for col in [c for c in enum_columns
-                    if c == self.order or c == self.group or c == 'priority']:
-            sql.append("\n  LEFT OUTER JOIN enum AS %s ON "
-                       "(%s.type='%s' AND %s.name=%s)"
-                       % (col, col, col, col, col))
+        sql.extend("\n  LEFT OUTER JOIN enum AS %(col)s ON "
+                   "(%(col)s.type='%(type)s' AND %(col)s.name=t.%(col)s)" %
+                   {'col': col,
+                    'type': 'ticket_type' if col == 'type' else col}
+                   for col in enum_columns)
 
         # Join with the version/milestone tables for proper sorting
-        for col in [c for c in ['milestone', 'version']
-                    if c == self.order or c == self.group]:
-            sql.append("\n  LEFT OUTER JOIN %s ON (%s.name=%s)"
-                       % (col, col, col))
+        sql.extend("\n  LEFT OUTER JOIN %(col)s ON (%(col)s.name=%(col)s)"
+                   % {'col': col} for col in joined_columns)
 
         def get_timestamp(date):
             if date:
@@ -668,12 +672,12 @@ class Query(object):
             if name in enum_columns:
                 # These values must be compared as ints, not as strings
                 sql.append(db.cast(col, 'int') + desc)
-            elif name == 'milestone':
+            elif name == 'milestone' and name not in custom_fields:
                 sql.append("COALESCE(milestone.completed,0)=0%s,"
                            "milestone.completed%s,"
                            "COALESCE(milestone.due,0)=0%s,milestone.due%s,"
                            "%s%s" % (desc, desc, desc, desc, col, desc))
-            elif name == 'version':
+            elif name == 'version' and name not in custom_fields:
                 sql.append("COALESCE(version.time,0)=0%s,version.time%s,%s%s"
                            % (desc, desc, col, desc))
             else:
@@ -1088,7 +1092,7 @@ class QueryModule(Component):
     def display_html(self, req, query):
         # The most recent query is stored in the user session;
         orig_list = None
-        orig_time = datetime.now(utc)
+        orig_time = datetime_now(utc)
         query_time = int(req.session.get('query_time', 0))
         query_time = datetime.fromtimestamp(query_time, utc)
         query_constraints = unicode(query.constraints)
@@ -1330,7 +1334,7 @@ class TicketQueryMacro(WikiMacroBase):
         clauses = [{}]
         argv = []
         kwargs = {}
-        for arg in TicketQueryMacro._comma_splitter.split(content):
+        for arg in TicketQueryMacro._comma_splitter.split(content or ''):
             arg = arg.replace(r'\,', ',')
             m = re.match(r'\s*[^=]+=', arg)
             if m:
@@ -1381,7 +1385,11 @@ class TicketQueryMacro(WikiMacroBase):
                              cnt, query=query_string)
             return tag.span(cnt, title=title, class_='query_count')
 
-        tickets = query.execute(req)
+        try:
+            tickets = query.execute(req)
+        except QueryValueError, e:
+            self.log.warn(e)
+            return system_message(_("Error executing TicketQuery macro"), e)
 
         if format == 'table':
             data = query.template_data(formatter.context, tickets,
