@@ -34,14 +34,12 @@ import shutil
 import sys
 import struct
 import tempfile
-import time
 from urllib import quote, unquote, urlencode
 
 from .compat import any, md5, sha1, sorted
-from .datefmt import to_datetime, to_timestamp, utc
+from .datefmt import time_now, to_datetime, to_timestamp, utc
 from .text import exception_to_unicode, to_unicode, getpreferredencoding
 
-# -- req, session and web utils
 
 def get_reporter_id(req, arg_name=None):
     """Get most informative "reporter" identity out of a request.
@@ -66,6 +64,7 @@ def get_reporter_id(req, arg_name=None):
         return '%s <%s>' % (name, email)
     return name or email or req.authname # == 'anonymous'
 
+
 def content_disposition(type=None, filename=None):
     """Generate a properly escaped Content-Disposition header."""
     type = type or ''
@@ -171,15 +170,15 @@ class AtomicFile(object):
     """
     def __init__(self, path, mode='w', bufsize=-1):
         self._file = None
-        self._path = path
-        (dir, name) = os.path.split(path)
-        (fd, self._temp) = tempfile.mkstemp(prefix=name + '-', dir=dir)
+        self._path = os.path.realpath(path)
+        dir, name = os.path.split(self._path)
+        fd, self._temp = tempfile.mkstemp(prefix=name + '-', dir=dir)
         self._file = os.fdopen(fd, mode, bufsize)
 
         # Try to preserve permissions and group ownership, but failure
         # should not be fatal
         try:
-            st = os.stat(path)
+            st = os.stat(self._path)
             if hasattr(os, 'chmod'):
                 os.chmod(self._temp, st.st_mode)
             if hasattr(os, 'chflags') and hasattr(st, 'st_flags'):
@@ -260,6 +259,27 @@ def create_unique_file(path):
             path = '%s.%d%s' % (parts[0], idx, parts[1])
 
 
+if os.name == 'nt':
+    def touch_file(filename):
+        """Update modified time of the given file. The file is created if
+        missing."""
+        # Use f.truncate() to avoid low resolution of GetSystemTime()
+        # on Windows
+        with open(filename, 'ab') as f:
+            stat = os.fstat(f.fileno())
+            f.truncate(stat.st_size)
+else:
+    def touch_file(filename):
+        """Update modified time of the given file. The file is created if
+        missing."""
+        try:
+            os.utime(filename, None)
+        except OSError, e:
+            if e.errno == errno.ENOENT:
+                with open(filename, 'ab'):
+                    pass
+
+
 def create_zipinfo(filename, mtime=None, dir=False, executable=False, symlink=False,
                    comment=None):
     """Create a instance of `ZipInfo`.
@@ -305,16 +325,16 @@ def create_zipinfo(filename, mtime=None, dir=False, executable=False, symlink=Fa
         if not zipinfo.filename.endswith('/'):
             zipinfo.filename += '/'
         zipinfo.compress_type = ZIP_STORED
-        zipinfo.external_attr = 040755 << 16L       # permissions drwxr-xr-x
-        zipinfo.external_attr |= 0x10               # MS-DOS directory flag
+        zipinfo.external_attr = 040755 << 16L        # permissions drwxr-xr-x
+        zipinfo.external_attr |= 0x10                # MS-DOS directory flag
     else:
         zipinfo.compress_type = ZIP_DEFLATED
-        zipinfo.external_attr = 0644 << 16L         # permissions -r-wr--r--
+        zipinfo.external_attr = 0644 << 16L          # permissions -r-wr--r--
         if executable:
-            zipinfo.external_attr |= 0755 << 16L    # -rwxr-xr-x
+            zipinfo.external_attr |= 0755 << 16L     # -rwxr-xr-x
         if symlink:
             zipinfo.compress_type = ZIP_STORED
-            zipinfo.external_attr |= 0120000 << 16L # symlink file type
+            zipinfo.external_attr |= 0120000 << 16L  # symlink file type
 
     if comment:
         zipinfo.comment = comment.encode('utf-8')
@@ -322,7 +342,7 @@ def create_zipinfo(filename, mtime=None, dir=False, executable=False, symlink=Fa
     return zipinfo
 
 
-class NaivePopen:
+class NaivePopen(object):
     """This is a deadlock-safe version of popen that returns an object with
     errorlevel, out (a string) and err (a string).
 
@@ -386,7 +406,6 @@ def terminate(process):
         except OSError, e:
             # If the process has already finished and has not been
             # waited for, killing it raises an ESRCH error on Cygwin
-            import errno
             if e.errno != errno.ESRCH:
                 raise
 
@@ -451,8 +470,8 @@ def copytree(src, dst, symlinks=False, skip=[], overwrite=False):
                 errors.extend(err.args[0])
         try:
             shutil.copystat(src, dst)
-        except WindowsError, why:
-            pass # Ignore errors due to limited Windows copystat support
+        except WindowsError:
+            pass  # Ignore errors due to limited Windows copystat support
         except OSError, why:
             errors.append((src, dst, str(why)))
         if errors:
@@ -520,7 +539,6 @@ def get_last_traceback():
 
 
 _egg_path_re = re.compile(r'build/bdist\.[^/]+/egg/(.*)')
-
 def get_lines_from_file(filename, lineno, context=0, globals=None):
     """Return `content` number of lines before and after the specified
     `lineno` from the (source code) file identified by `filename`.
@@ -564,10 +582,10 @@ def get_lines_from_file(filename, lineno, context=0, globals=None):
             break
 
     before = [to_unicode(l.rstrip('\n'), charset)
-                 for l in lines[lbound:lineno]]
+              for l in lines[lbound:lineno]]
     line = to_unicode(lines[lineno].rstrip('\n'), charset)
-    after = [to_unicode(l.rstrip('\n'), charset) \
-                 for l in lines[lineno + 1:ubound]]
+    after = [to_unicode(l.rstrip('\n'), charset)
+             for l in lines[lineno + 1:ubound]]
 
     return before, line, after
 
@@ -637,15 +655,14 @@ def get_doc(obj):
     """
     doc = inspect.getdoc(obj)
     if not doc:
-        return (None, None)
+        return None, None
     doc = to_unicode(doc).split('\n\n', 1)
     summary = doc[0].replace('\n', ' ')
     description = doc[1] if len(doc) > 1 else None
-    return (summary, description)
+    return summary, description
 
 
 _dont_import = frozenset(['__file__', '__name__', '__package__'])
-
 def import_namespace(globals_dict, module_name):
     """Import the namespace of a module into a globals dict.
 
@@ -678,23 +695,31 @@ def get_module_path(module):
             break
     return base_path
 
+
 def get_sources(path):
     """Return a dictionary mapping Python module source paths to the
     distributions that contain them.
     """
     sources = {}
     for dist in find_distributions(path, only=True):
-        try:
-            toplevels = dist.get_metadata('top_level.txt').splitlines()
-            toplevels = [each + '/' for each in toplevels]
-            files = dist.get_metadata('SOURCES.txt').splitlines()
-            sources.update((src, dist) for src in files
-                           if any(src.startswith(toplevel)
-                                  for toplevel in toplevels))
-        except (KeyError, IOError):
-            pass    # Metadata not found
+        if not dist.has_metadata('top_level.txt'):
+            continue
+        toplevels = dist.get_metadata_lines('top_level.txt')
+        toplevels = [top + '/' for top in toplevels]
+        if dist.has_metadata('SOURCES.txt'):  # *.egg-info/SOURCES.txt
+            sources.update((src, dist)
+                           for src in dist.get_metadata_lines('SOURCES.txt')
+                           if any(src.startswith(top) for top in toplevels))
+            continue
+        if dist.has_metadata('RECORD'):  # *.dist-info/RECORD
+            reader = csv.reader(StringIO(dist.get_metadata('RECORD')))
+            sources.update((row[0], dist)
+                           for row in reader if any(row[0].startswith(top)
+                                                    for top in toplevels))
+            continue
     return sources
 
+
 def get_pkginfo(dist):
     """Get a dictionary containing package information for a package
 
@@ -789,7 +814,7 @@ except NotImplementedError:
 
     def urandom(n):
         result = []
-        hasher = sha1(str(os.getpid()) + str(time.time()))
+        hasher = sha1(str(os.getpid()) + str(time_now()))
         while len(result) * hasher.digest_size < n:
             hasher.update(str(_entropy.random()))
             result.append(hasher.digest())
@@ -802,6 +827,7 @@ def hex_entropy(digits=32):
     result = ''.join('%.2x' % ord(v) for v in urandom((digits + 1) // 2))
     return result[:digits] if len(result) > digits else result
 
+
 # Original license for md5crypt:
 # Based on FreeBSD src/lib/libcrypt/crypt.c 1.2
 #
@@ -990,15 +1016,15 @@ class Ranges(object):
         p.sort()
         i = 0
         while i + 1 < len(p):
-            if p[i+1][0]-1 <= p[i][1]: # this item overlaps with the next
+            if p[i+1][0]-1 <= p[i][1]:  # this item overlaps with the next
                 # make the first include the second
                 p[i] = (p[i][0], max(p[i][1], p[i+1][1]))
-                del p[i+1] # delete the second, after adjusting my endpoint
+                del p[i+1]  # delete the second, after adjusting my endpoint
             else:
                 i += 1
         if p:
-            self.a = p[0][0] # min value
-            self.b = p[-1][1] # max value
+            self.a = p[0][0]   # min value
+            self.b = p[-1][1]  # max value
         else:
             self.a = self.b = None
 
@@ -1157,6 +1183,7 @@ def embedded_numbers(s):
     pieces[1::2] = map(int, pieces[1::2])
     return pieces
 
+
 def pairwise(iterable):
     """
     >>> list(pairwise([0, 1, 2, 3]))
@@ -1172,6 +1199,7 @@ def pairwise(iterable):
         pass
     return izip(a, b)
 
+
 def partition(iterable, order=None):
     """
     >>> partition([(1, "a"), (2, "b"), (3, "a")])
@@ -1189,6 +1217,7 @@ def partition(iterable, order=None):
         return result
     return [result[key] for key in order]
 
+
 def as_int(s, default, min=None, max=None):
     """Convert s to an int and limit it to the given range, or return default
     if unsuccessful."""
@@ -1202,6 +1231,7 @@ def as_int(s, default, min=None, max=None):
         value = max
     return value
 
+
 def as_bool(value):
     """Convert the given value to a `bool`.
 
@@ -1219,6 +1249,7 @@ def as_bool(value):
     except (TypeError, ValueError):
         return False
 
+
 def pathjoin(*args):
     """Strip `/` from the arguments and join them with a single `/`."""
     return '/'.join(filter(None, (each.strip('/') for each in args if each)))
