@@ -1,8 +1,8 @@
 # -*- coding: utf-8 -*-
 #
-# Copyright (C) 2003-2009 Edgewall Software
+# Copyright (C) 2003-2010 Edgewall Software
 # Copyright (C) 2003-2005 Jonas Borgström <jonas@edgewall.com>
-# Copyright (C) 2005-2007 Christian Boos <cboos@neuf.fr>
+# Copyright (C) 2005-2007 Christian Boos <cboos@edgewall.org>
 # All rights reserved.
 #
 # This software is licensed as described in the file COPYING, which
@@ -23,23 +23,23 @@ from genshi.builder import tag
 
 from trac.config import ListOption, BoolOption, Option
 from trac.core import *
-from trac.mimeview.api import Mimeview, is_binary, \
-                              IHTMLPreviewAnnotator, Context
-from trac.perm import IPermissionRequestor
+from trac.mimeview.api import IHTMLPreviewAnnotator, Mimeview, is_binary
+from trac.perm import IPermissionRequestor, PermissionError
 from trac.resource import Resource, ResourceNotFound
 from trac.util import as_bool, embedded_numbers
-from trac.util.compat import any, cleandoc
-from trac.util.datefmt import http_date, to_datetime, utc
+from trac.util.datefmt import datetime_now, http_date, to_datetime, utc
 from trac.util.html import escape, Markup
 from trac.util.text import exception_to_unicode, shorten_line
-from trac.util.translation import _
-from trac.web import IRequestHandler, RequestDone
-from trac.web.chrome import add_ctxtnav, add_link, add_script, add_stylesheet, \
-                            prevnext_nav, INavigationContributor
+from trac.util.translation import _, cleandoc_
+from trac.web.api import IRequestHandler, RequestDone
+from trac.web.chrome import (INavigationContributor, add_ctxtnav, add_link,
+                             add_script, add_stylesheet, prevnext_nav,
+                             web_context)
 from trac.wiki.api import IWikiSyntaxProvider, IWikiMacroProvider, parse_args
 from trac.wiki.formatter import format_to_html, format_to_oneliner
-from trac.versioncontrol.api import NoSuchChangeset, RepositoryManager
-from trac.versioncontrol.web_ui.util import *
+
+from ..api import NoSuchChangeset, RepositoryManager
+from trac.versioncontrol.web_ui.util import * # `from .util import *` FIXME 2.6
 
 
 CHUNK_SIZE = 4096
@@ -129,8 +129,8 @@ class WikiPropertyRenderer(Component):
         (''since 0.11'')""")
 
     def match_property(self, name, mode):
-        return (name in self.wiki_properties or \
-                name in self.oneliner_properties) and 4 or 0
+        return 4 if name in self.wiki_properties \
+                    or name in self.oneliner_properties else 0
 
     def render_property(self, name, mode, context, props):
         if name in self.wiki_properties:
@@ -142,7 +142,7 @@ class WikiPropertyRenderer(Component):
 class TimeRange(object):
 
     min = datetime(1, 1, 1, 0, 0, 0, 0, utc) # tz aware version of datetime.min
-    
+
     def __init__(self, base):
         self.oldest = self.newest = base
         self._total = None
@@ -150,7 +150,7 @@ class TimeRange(object):
     def seconds_between(self, dt1, dt2):
         delta = dt1 - dt2
         return delta.days * 24 * 3600 + delta.seconds
-    
+
     def to_seconds(self, dt):
         return self.seconds_between(dt, TimeRange.min)
 
@@ -175,7 +175,7 @@ class TimeRange(object):
 class BrowserModule(Component):
 
     implements(INavigationContributor, IPermissionRequestor, IRequestHandler,
-               IWikiSyntaxProvider, IHTMLPreviewAnnotator, 
+               IWikiSyntaxProvider, IHTMLPreviewAnnotator,
                IWikiMacroProvider)
 
     property_renderers = ExtensionPoint(IPropertyRenderer)
@@ -183,15 +183,19 @@ class BrowserModule(Component):
     downloadable_paths = ListOption('browser', 'downloadable_paths',
                                     '/trunk, /branches/*, /tags/*',
         doc="""List of repository paths that can be downloaded.
-        
-        Leave the option empty if you want to disable all downloads, otherwise
+
+        Leave this option empty if you want to disable all downloads, otherwise
         set it to a comma-separated list of authorized paths (those paths are
-        glob patterns, i.e. "*" can be used as a wild card)
+        glob patterns, i.e. "*" can be used as a wild card). In a
+        multi-repository environment, the path must be qualified with the
+        repository name if the path does not point to the default repository
+        (e.g. /reponame/trunk). Note that a simple prefix matching is
+        performed on the paths, so aliases won't get automatically resolved.
         (''since 0.10'')""")
 
     color_scale = BoolOption('browser', 'color_scale', True,
         doc="""Enable colorization of the ''age'' column.
-        
+
         This uses the same color scale as the source code annotation:
         blue is older, red is newer.
         (''since 0.11'')""")
@@ -228,13 +232,13 @@ class BrowserModule(Component):
 
     render_unsafe_content = BoolOption('browser', 'render_unsafe_content',
                                         'false',
-        """Whether raw files should be rendered in the browser, or only made 
+        """Whether raw files should be rendered in the browser, or only made
         downloadable.
- 
+
         Pretty much any file may be interpreted as HTML by the browser,
         which allows a malicious user to create a file containing cross-site
         scripting attacks.
-        
+
         For open repositories where anyone can check-in a file, it is
         recommended to leave this option disabled (which is the default).""")
 
@@ -247,7 +251,7 @@ class BrowserModule(Component):
 
     def get_custom_colorizer(self):
         """Returns a converter for values from [0.0, 1.0] to a RGB triple."""
-        
+
         def interpolate(old, new, value):
             # Provides a linearly interpolated color triple for `value`
             # which must be a floating point value between 0.0 and 1.0
@@ -257,10 +261,10 @@ class BrowserModule(Component):
             # Get three ints out of a `rgb` string or return `default`
             try:
                 t = tuple([int(v) for v in re.split(r'(\d+)', rgb)[1::2]])
-                return len(t) == 3 and t or default
+                return t if len(t) == 3 else default
             except ValueError:
                 return default
-        
+
         newest_color = parse_color(self.newest_color, self.NEWEST_COLOR)
         oldest_color = parse_color(self.oldest_color, self.OLDEST_COLOR)
         try:
@@ -283,7 +287,7 @@ class BrowserModule(Component):
             def colorizer(value):
                 return interpolate(oldest_color, newest_color, value)
         return colorizer
-        
+
     # INavigationContributor methods
 
     def get_active_navigation_item(self, req):
@@ -291,7 +295,8 @@ class BrowserModule(Component):
 
     def get_navigation_items(self, req):
         rm = RepositoryManager(self.env)
-        if 'BROWSER_VIEW' in req.perm and rm.get_real_repositories():
+        if any(repos.is_viewable(req.perm) for repos
+                                           in rm.get_real_repositories()):
             yield ('mainnav', 'browser',
                    tag.a(_('Browse Source'), href=req.href.browser()))
 
@@ -322,20 +327,19 @@ class BrowserModule(Component):
             return True
 
     def process_request(self, req):
-        req.perm.require('BROWSER_VIEW')
-
         presel = req.args.get('preselected')
         if presel and (presel + '/').startswith(req.href.browser() + '/'):
             req.redirect(presel)
 
         path = req.args.get('path', '/')
         rev = req.args.get('rev', '')
-        if rev in ('', 'HEAD'):
+        if rev.lower() in ('', 'head'):
             rev = None
+        format = req.args.get('format')
         order = req.args.get('order', 'name').lower()
-        desc = req.args.has_key('desc')
+        desc = 'desc' in req.args
         xhr = req.get_header('X-Requested-With') == 'XMLHttpRequest'
-        
+
         rm = RepositoryManager(self.env)
         all_repositories = rm.get_all_repositories()
         reponame, repos, path = rm.get_repository_by_path(path)
@@ -344,7 +348,7 @@ class BrowserModule(Component):
         show_index = not reponame and path == '/'
         if show_index:
             if repos and (as_bool(all_repositories[''].get('hidden'))
-                          or not repos.can_view(req.perm)):
+                          or not repos.is_viewable(req.perm)):
                 repos = None
 
         if not repos and reponame:
@@ -354,12 +358,13 @@ class BrowserModule(Component):
         if reponame and reponame != repos.reponame: # Redirect alias
             qs = req.query_string
             req.redirect(req.href.browser(repos.reponame or None, path)
-                         + (qs and '?' + qs or ''))
-        reponame = repos and repos.reponame or None
-        
+                         + ('?' + qs if qs else ''))
+        reponame = repos.reponame if repos else None
+
         # Find node for the requested path/rev
-        context = Context.from_request(req)
+        context = web_context(req)
         node = None
+        changeset = None
         display_rev = lambda rev: rev
         if repos:
             try:
@@ -370,10 +375,15 @@ class BrowserModule(Component):
                 rev_or_latest = rev or repos.youngest_rev
                 node = get_existing_node(req, repos, path, rev_or_latest)
             except NoSuchChangeset, e:
-                raise ResourceNotFound(e.message,
-                                       _('Invalid changeset number'))
+                raise ResourceNotFound(e, _('Invalid changeset number'))
+            if node:
+                try:
+                    # use changeset instance to retrieve branches and tags
+                    changeset = repos.get_changeset(node.rev)
+                except NoSuchChangeset:
+                    pass
 
-            context = context(repos.resource.child('source', path,
+            context = context.child(repos.resource.child('source', path,
                                                    version=rev_or_latest))
             display_rev = repos.display_rev
 
@@ -386,13 +396,25 @@ class BrowserModule(Component):
             repo_data = self._render_repository_index(
                                         context, all_repositories, order, desc)
         if node:
+            if not node.is_viewable(req.perm):
+                raise PermissionError('BROWSER_VIEW' if node.isdir else
+                                      'FILE_VIEW', node.resource, self.env)
             if node.isdir:
+                if format in ('zip',): # extension point here...
+                    self._render_zip(req, context, repos, node, rev)
+                    # not reached
                 dir_data = self._render_dir(req, repos, node, rev, order, desc)
             elif node.isfile:
                 file_data = self._render_file(req, context, repos, node, rev)
 
         if not repos and not (repo_data and repo_data['repositories']):
-            raise ResourceNotFound(_("No node %(path)s", path=path))
+            # If no viewable repositories, check permission instead of
+            # repos.is_viewable()
+            req.perm.require('BROWSER_VIEW')
+            if show_index:
+                raise ResourceNotFound(_("No viewable repositories"))
+            else:
+                raise ResourceNotFound(_("No node %(path)s", path=path))
 
         quickjump_data = properties_data = None
         if node and not xhr:
@@ -404,18 +426,18 @@ class BrowserModule(Component):
             'context': context, 'reponame': reponame, 'repos': repos,
             'repoinfo': all_repositories.get(reponame or ''),
             'path': path, 'rev': node and node.rev, 'stickyrev': rev,
-            'display_rev': display_rev,
+            'display_rev': display_rev, 'changeset': changeset,
             'created_path': node and node.created_path,
             'created_rev': node and node.created_rev,
             'properties': properties_data,
             'path_links': path_links,
-            'order': order, 'desc': desc and 1 or None,
+            'order': order, 'desc': 1 if desc else None,
             'repo': repo_data, 'dir': dir_data, 'file': file_data,
             'quickjump_entries': quickjump_data,
             'wiki_format_messages': \
                 self.config['changeset'].getbool('wiki_format_messages'),
             'xhr': xhr,
-        }                   
+        }
         if xhr: # render and return the content only
             return 'dir_entries.html', data, None
 
@@ -449,27 +471,27 @@ class BrowserModule(Component):
                 if path != '/':
                     add_link(req, 'up', path_links[-2]['href'],
                              _('Parent directory'))
-                add_ctxtnav(req, tag.a(_('Last Change'), 
+                add_ctxtnav(req, tag.a(_('Last Change'),
                             href=req.href.changeset(node.created_rev, reponame,
                                                     node.created_path)))
             if node.isfile:
                 annotate = data['file']['annotate']
                 if annotate:
-                    add_ctxtnav(req, _('Normal'), 
-                                title=_('View file without annotations'), 
+                    add_ctxtnav(req, _('Normal'),
+                                title=_('View file without annotations'),
                                 href=req.href.browser(reponame,
-                                                      node.created_path, 
+                                                      node.created_path,
                                                       rev=rev))
                 if annotate != 'blame':
-                    add_ctxtnav(req, _('Blame'), 
+                    add_ctxtnav(req, _('Blame'),
                                 title=_('Annotate each line with the last '
                                         'changed revision '
-                                        '(this can be time consuming...)'), 
+                                        '(this can be time consuming...)'),
                                 href=req.href.browser(reponame,
-                                                      node.created_path, 
+                                                      node.created_path,
                                                       rev=rev,
                                                       annotate='blame'))
-            add_ctxtnav(req, _('Revision Log'), 
+            add_ctxtnav(req, _('Revision Log'),
                         href=req.href.log(reponame, path, rev=rev))
             path_url = repos.get_path_url(path, rev)
             if path_url:
@@ -495,8 +517,12 @@ class BrowserModule(Component):
                 continue
             try:
                 repos = rm.get_repository(reponame)
+            except TracError, err:
+                entry = (reponame, repoinfo, None, None,
+                         exception_to_unicode(err), None)
+            else:
                 if repos:
-                    if not repos.can_view(context.perm):
+                    if not repos.is_viewable(context.perm):
                         continue
                     try:
                         youngest = repos.get_changeset(repos.youngest_rev)
@@ -507,13 +533,13 @@ class BrowserModule(Component):
                             timerange = TimeRange(youngest.date)
                         else:
                             timerange.insert(youngest.date)
-                    entry = (reponame, repoinfo, repos, youngest, None)
+                    raw_href = self._get_download_href(context.href, repos,
+                                                       None, None)
+                    entry = (reponame, repoinfo, repos, youngest, None,
+                             raw_href)
                 else:
-                    entry = (reponame, repoinfo, None, None, u"\u2013")
-            except TracError, err:
-                entry = (reponame, repoinfo, None, None,
-                         exception_to_unicode(err))
-            if entry[-1] is not None:   # Check permission in case of error
+                    entry = (reponame, repoinfo, None, None, u"\u2013", None)
+            if entry[4] is not None:  # Check permission in case of error
                 root = Resource('repository', reponame).child('source', '/')
                 if 'BROWSER_VIEW' not in context.perm(root):
                     continue
@@ -521,15 +547,15 @@ class BrowserModule(Component):
 
         # Ordering of repositories
         if order == 'date':
-            def repo_order((reponame, repoinfo, repos, youngest, err)):
-                return (youngest and youngest.date or to_datetime(0),
+            def repo_order((reponame, repoinfo, repos, youngest, err, href)):
+                return (youngest.date if youngest else to_datetime(0),
                         embedded_numbers(reponame.lower()))
         elif order == 'author':
-            def repo_order((reponame, repoinfo, repos, youngest, err)):
-                return (youngest and youngest.author.lower() or '',
+            def repo_order((reponame, repoinfo, repos, youngest, err, href)):
+                return (youngest.author.lower() if youngest else '',
                         embedded_numbers(reponame.lower()))
         else:
-            def repo_order((reponame, repoinfo, repos, youngest, err)):
+            def repo_order((reponame, repoinfo, repos, youngest, err, href)):
                 return embedded_numbers(reponame.lower())
 
         repositories = sorted(repositories, key=repo_order, reverse=desc)
@@ -539,25 +565,28 @@ class BrowserModule(Component):
 
     def _render_dir(self, req, repos, node, rev, order, desc):
         req.perm(node.resource).require('BROWSER_VIEW')
+        download_href = self._get_download_href
 
         # Entries metadata
         class entry(object):
-            __slots__ = ('name', 'rev', 'created_rev', 'kind', 'isdir', 'path',
-                         'content_length')
+            _copy = 'name rev created_rev kind isdir path content_length' \
+                    .split()
+            __slots__ = _copy + ['raw_href']
 
             def __init__(self, node):
-                for f in entry.__slots__:
+                for f in entry._copy:
                     setattr(self, f, getattr(node, f))
-                
+                self.raw_href = download_href(req.href, repos, node, rev)
+
         entries = [entry(n) for n in node.get_entries()
-                   if n.can_view(req.perm)]
+                   if n.is_viewable(req.perm)]
         changes = get_changes(repos, [i.created_rev for i in entries],
                               self.log)
 
         if rev:
             newest = repos.get_changeset(rev).date
         else:
-            newest = datetime.now(req.tz)
+            newest = datetime_now(req.tz)
 
         # Color scale for the age column
         timerange = custom_colorizer = None
@@ -589,23 +618,15 @@ class BrowserModule(Component):
             def file_order(a):
                 return embedded_numbers(a.name.lower())
 
-        dir_order = desc and 1 or -1
+        dir_order = 1 if desc else -1
 
         def browse_order(a):
-            return a.isdir and dir_order or 0, file_order(a)
+            return dir_order if a.isdir else 0, file_order(a)
         entries = sorted(entries, key=browse_order, reverse=desc)
 
         # ''Zip Archive'' alternate link
-        path = node.path.strip('/')
-        if repos.reponame:
-            path = repos.reponame + '/' + path
-        if any(fnmatchcase(path, p.strip('/'))
-               for p in self.downloadable_paths):
-            zip_href = req.href.changeset(rev or repos.youngest_rev, 
-                                          repos.reponame or None, node.path,
-                                          old=rev,
-                                          old_path=repos.reponame or '/',
-                                          format='zip')
+        zip_href = self._get_download_href(req.href, repos, node, rev)
+        if zip_href:
             add_link(req, 'alternate', zip_href, _('Zip Archive'),
                      'application/zip', 'zip')
 
@@ -617,13 +638,35 @@ class BrowserModule(Component):
                                    timerange.to_seconds(timerange.oldest)),
                 }
 
+    def _iter_nodes(self, node):
+        stack = [node]
+        while stack:
+            node = stack.pop()
+            yield node
+            if node.isdir:
+                stack.extend(sorted(node.get_entries(),
+                                    key=lambda x: x.name,
+                                    reverse=True))
+
+    def _render_zip(self, req, context, repos, root_node, rev=None):
+        if not self.is_path_downloadable(repos, root_node.path):
+            raise TracError(_("Path not available for download"))
+        req.perm(context.resource).require('FILE_VIEW')
+        root_path = root_node.path.rstrip('/')
+        if root_path:
+            archive_name = root_node.name
+        else:
+            archive_name = repos.reponame or 'repository'
+        filename = '%s-%s.zip' % (archive_name, root_node.rev)
+        render_zip(req, filename, repos, root_node, self._iter_nodes)
+
     def _render_file(self, req, context, repos, node, rev=None):
         req.perm(node.resource).require('FILE_VIEW')
 
         mimeview = Mimeview(self.env)
 
         # MIME type detection
-        content = node.get_content()
+        content = node.get_processed_content()
         chunk = content.read(CHUNK_SIZE)
         mime_type = node.content_type
         if not mime_type or mime_type == 'application/octet-stream':
@@ -635,8 +678,7 @@ class BrowserModule(Component):
         if format in ('raw', 'txt'):
             req.send_response(200)
             req.send_header('Content-Type',
-                            format == 'txt' and 'text/plain' or mime_type)
-            req.send_header('Content-Length', node.content_length)
+                            'text/plain' if format == 'txt' else mime_type)
             req.send_header('Last-Modified', http_date(node.last_modified))
             if rev is None:
                 req.send_header('Pragma', 'no-cache')
@@ -644,18 +686,18 @@ class BrowserModule(Component):
                 req.send_header('Expires', 'Fri, 01 Jan 1999 00:00:00 GMT')
             if not self.render_unsafe_content:
                 # Force browser to download files instead of rendering
-                # them, since they might contain malicious code enabling 
+                # them, since they might contain malicious code enabling
                 # XSS attacks
                 req.send_header('Content-Disposition', 'attachment')
             req.end_headers()
-
-            while 1:
-                if not chunk:
-                    raise RequestDone
+            # Note: don't pass an iterable instance to RequestDone, instead
+            # call req.write() with each chunk here to avoid SEGVs (#11805)
+            while chunk:
                 req.write(chunk)
                 chunk = content.read(CHUNK_SIZE)
+            raise RequestDone
         else:
-            # The changeset corresponding to the last change on `node` 
+            # The changeset corresponding to the last change on `node`
             # is more interesting than the `rev` changeset.
             changeset = repos.get_changeset(node.created_rev)
 
@@ -667,15 +709,15 @@ class BrowserModule(Component):
                          'text/plain')
 
             # add ''Original Format'' alternate link (always)
-            raw_href = req.href.export(rev or repos.youngest_rev, 
+            raw_href = req.href.export(rev or repos.youngest_rev,
                                        repos.reponame or None, node.path)
             add_link(req, 'alternate', raw_href, _('Original Format'),
                      mime_type)
 
-            self.log.debug("Rendering preview of node %s@%s with mime-type %s"
-                           % (node.name, str(rev), mime_type))
+            self.log.debug("Rendering preview of node %s@%s with mime-type %s",
+                           node.name, rev, mime_type)
 
-            del content # the remainder of that content is not needed
+            content = None # the remainder of that content is not needed
 
             add_stylesheet(req, 'common/css/code.css')
 
@@ -683,7 +725,8 @@ class BrowserModule(Component):
             annotate = req.args.get('annotate')
             if annotate:
                 annotations.insert(0, annotate)
-            preview_data = mimeview.preview_data(context, node.get_content(),
+            preview_data = mimeview.preview_data(context,
+                                                 node.get_processed_content(),
                                                  node.get_content_length(),
                                                  mime_type, node.created_path,
                                                  raw_href,
@@ -696,8 +739,24 @@ class BrowserModule(Component):
                 'annotate': annotate,
                 }
 
+    def _get_download_href(self, href, repos, node, rev):
+        """Return the URL for downloading a file, or a directory as a ZIP."""
+        if node is not None and node.isfile:
+            return href.export(rev or 'HEAD', repos.reponame or None,
+                               node.path)
+        path = '' if node is None else node.path.strip('/')
+        if self.is_path_downloadable(repos, path):
+            return href.browser(repos.reponame or None, path,
+                                rev=rev or repos.youngest_rev, format='zip')
+
     # public methods
-    
+
+    def is_path_downloadable(self, repos, path):
+        if repos.reponame:
+            path = repos.reponame + '/' + path
+        return any(fnmatchcase(path, dp.strip('/'))
+                   for dp in self.downloadable_paths)
+
     def render_properties(self, mode, context, props):
         """Prepare rendering of a collection of properties."""
         return filter(None, [self.render_property(name, mode, context, props)
@@ -760,9 +819,13 @@ class BrowserModule(Component):
         elif '@' in export:
             path, rev = export.split('@', 1)
         else:
-            rev, path = 'HEAD', export
-        return tag.a(label, class_='export',
-                     href=formatter.href.export(rev, path) + fragment)
+            rev, path = None, export
+        node, raw_href, title = self._get_link_info(path, rev, formatter.href,
+                                                    formatter.perm)
+        if raw_href:
+            return tag.a(label, class_='export', href=raw_href + fragment,
+                         title=title)
+        return tag.a(label, class_='missing export')
 
     def _format_browser_link(self, formatter, ns, path, label):
         path, query, fragment = formatter.split_link(path)
@@ -770,15 +833,39 @@ class BrowserModule(Component):
         match = self.PATH_LINK_RE.match(path)
         if match:
             path, rev, marks = match.groups()
-        return tag.a(label, class_='source',
-                     href=(formatter.href.browser(path, rev=rev, marks=marks) +
-                           query + fragment))
+        href = formatter.href
+        src_href = href.browser(path, rev=rev, marks=marks) + query + fragment
+        node, raw_href, title = self._get_link_info(path, rev, formatter.href,
+                                                    formatter.perm)
+        if not node:
+            return tag.a(label, class_='missing source')
+        link = tag.a(label, class_='source', href=src_href)
+        if raw_href:
+            link = tag(link, tag.a(u'\u200b', href=raw_href + fragment,
+                                   title=title,
+                                   class_='trac-rawlink' if node.isfile
+                                          else 'trac-ziplink'))
+        return link
 
     PATH_LINK_RE = re.compile(r"([^@#:]*)"     # path
                               r"[@:]([^#:]+)?" # rev
                               r"(?::(\d+(?:-\d+)?(?:,\d+(?:-\d+)?)*))?" # marks
                               )
 
+    def _get_link_info(self, path, rev, href, perm):
+        rm = RepositoryManager(self.env)
+        node = raw_href = title = None
+        try:
+            reponame, repos, npath = rm.get_repository_by_path(path)
+            node = get_allowed_node(repos, npath, rev, perm)
+            if node is not None:
+                raw_href = self._get_download_href(href, repos, node, rev)
+                title = _("Download") if node.isfile \
+                        else _("Download as Zip archive")
+        except TracError:
+            pass
+        return (node, raw_href, title)
+
     # IHTMLPreviewAnnotator methods
 
     def get_annotation_type(self):
@@ -797,16 +884,16 @@ class BrowserModule(Component):
         yield "RepositoryIndex"
 
     def get_macro_description(self, name):
-        return cleandoc("""
+        description = cleandoc_("""
         Display the list of available repositories.
 
         Can be given the following named arguments:
-        
+
           ''format''::
             Select the rendering format:
             - ''compact'' produces a comma-separated list of repository prefix
               names (default)
-            - ''list'' produces a description list of repository prefix names 
+            - ''list'' produces a description list of repository prefix names
             - ''table'' produces a table view, similar to the one visible in
               the ''Browse View'' page
           ''glob''::
@@ -819,6 +906,7 @@ class BrowserModule(Component):
 
         (''since 0.12'')
         """)
+        return 'messages', description
 
     def expand_macro(self, formatter, name, content):
         args, kwargs = parse_args(content)
@@ -838,7 +926,7 @@ class BrowserModule(Component):
             add_stylesheet(formatter.req, 'common/css/browser.css')
             wiki_format_messages = self.config['changeset'] \
                                        .getbool('wiki_format_messages')
-            data = {'repo': repo, 'order': order, 'desc': desc and 1 or None,
+            data = {'repo': repo, 'order': order, 'desc': 1 if desc else None,
                     'reponame': None, 'path': '/', 'stickyrev': None,
                     'wiki_format_messages': wiki_format_messages}
             from trac.web.chrome import Chrome
@@ -857,12 +945,12 @@ class BrowserModule(Component):
         all_repos = sorted(((reponame, repos) for reponame, repos in all_repos
                             if repos
                             and not as_bool(repos.params.get('hidden'))
-                            and repos.can_view(formatter.perm)),
+                            and repos.is_viewable(formatter.perm)),
                            reverse=desc)
 
         def repolink(reponame, repos):
             label = reponame or _('(default)')
-            return Markup(tag.a(label, 
+            return Markup(tag.a(label,
                           title=_('View repository %(repo)s', repo=label),
                           href=formatter.href.browser(repos.reponame or None)))
 
@@ -875,7 +963,7 @@ class BrowserModule(Component):
             return Markup(', ').join([repolink(reponame, repos)
                                       for reponame, repos in all_repos])
 
-        
+
 
 class BlameAnnotator(object):
 
@@ -935,7 +1023,7 @@ class BlameAnnotator(object):
         path = self.paths.get(rev, None)
         # Note: path will be None if copy/rename is not supported
         # by get_history
-        
+
         # -- compute anchor and style once per revision
         if rev not in self.chgset_data:
             chgset_href = \
