@@ -2,7 +2,7 @@
 #
 # Copyright (C) 2003-2009 Edgewall Software
 # Copyright (C) 2003-2005 Jonas Borgström <jonas@edgewall.com>
-# Copyright (C) 2005-2006 Christian Boos <cboos@neuf.fr>
+# Copyright (C) 2005-2006 Christian Boos <cboos@edgewall.org>
 # All rights reserved.
 #
 # This software is licensed as described in the file COPYING, which
@@ -14,30 +14,30 @@
 # history and logs, available at http://trac.edgewall.org/log/.
 #
 # Author: Jonas Borgström <jonas@edgewall.com>
-#         Christian Boos <cboos@neuf.fr>
+#         Christian Boos <cboos@edgewall.org>
 
 import re
 
 from genshi.core import Markup
 from genshi.builder import tag
 
-from trac.config import IntOption
+from trac.config import IntOption, ListOption
 from trac.core import *
-from trac.mimeview import Context
 from trac.perm import IPermissionRequestor
 from trac.resource import ResourceNotFound
-from trac.util import Ranges
-from trac.util.compat import any
+from trac.util import Ranges, as_int
 from trac.util.text import to_unicode, wrap
 from trac.util.translation import _
-from trac.versioncontrol.api import RepositoryManager, Changeset, \
-                                    NoSuchChangeset
+from trac.versioncontrol.api import (Changeset, NoSuchChangeset,
+                                     RepositoryManager)
 from trac.versioncontrol.web_ui.changeset import ChangesetModule
 from trac.versioncontrol.web_ui.util import *
 from trac.web import IRequestHandler
-from trac.web.chrome import add_ctxtnav, add_link, add_stylesheet, \
-                            INavigationContributor, Chrome
-from trac.wiki import IWikiSyntaxProvider, WikiParser 
+from trac.web.chrome import (Chrome, INavigationContributor, add_ctxtnav,
+                             add_link, add_script, add_script_data,
+                             add_stylesheet, auth_link, web_context)
+from trac.wiki import IWikiSyntaxProvider, WikiParser
+
 
 class LogModule(Component):
 
@@ -45,8 +45,13 @@ class LogModule(Component):
                IWikiSyntaxProvider)
 
     default_log_limit = IntOption('revisionlog', 'default_log_limit', 100,
-        """Default value for the limit argument in the TracRevisionLog
-        (''since 0.11'').""")
+        """Default value for the limit argument in the TracRevisionLog.
+        (''since 0.11'')""")
+
+    graph_colors = ListOption('revisionlog', 'graph_colors',
+        ['#cc0', '#0c0', '#0cc', '#00c', '#c0c', '#c00'],
+        doc="""Comma-separated list of colors to use for the TracRevisionLog
+        graph display. (''since 1.0'')""")
 
     # INavigationContributor methods
 
@@ -79,77 +84,75 @@ class LogModule(Component):
         revs = req.args.get('revs')
         format = req.args.get('format')
         verbose = req.args.get('verbose')
-        limit = int(req.args.get('limit') or self.default_log_limit)
+        limit = as_int(req.args.get('limit'), self.default_log_limit)
 
         rm = RepositoryManager(self.env)
         reponame, repos, path = rm.get_repository_by_path(path)
-        
+
         if not repos:
-            raise ResourceNotFound(_("Repository '%(repo)s' not found",
-                                   repo=reponame))
+            if path == '/':
+                raise TracError(_("No repository specified and no default"
+                                  " repository configured."))
+            else:
+                raise ResourceNotFound(_("Repository '%(repo)s' not found",
+                                         repo=reponame or path.strip('/')))
 
         if reponame != repos.reponame:  # Redirect alias
             qs = req.query_string
             req.redirect(req.href.log(repos.reponame or None, path)
-                         + (qs and '?' + qs or ''))
+                         + ('?' + qs if qs else ''))
 
         normpath = repos.normalize_path(path)
-        # if `revs` parameter is given, then we're restricted to the 
+
+        # if `revs` parameter is given, then we're restricted to the
         # corresponding revision ranges.
-        # If not, then we're considering all revisions since `rev`, 
+        # If not, then we're considering all revisions since `rev`,
         # on that path, in which case `revranges` will be None.
-        revranges = None
         if revs:
-            try:
-                revranges = Ranges(revs)
-                rev = revranges.b
-            except ValueError:
-                pass
-        rev = unicode(repos.normalize_rev(rev))
-        display_rev = repos.display_rev
+            revranges = RevRanges(repos, revs, resolve=True)
+            rev = revranges.b
+        else:
+            revranges = None
+            rev = repos.normalize_rev(rev)
 
         # The `history()` method depends on the mode:
         #  * for ''stop on copy'' and ''follow copies'', it's `Node.history()`
         #    unless explicit ranges have been specified
         #  * for ''show only add, delete'' we're using
-        #   `Repository.get_path_history()` 
+        #   `Repository.get_path_history()`
         cset_resource = repos.resource.child('changeset')
+        show_graph = False
+        curr_revrange = []
         if mode == 'path_history':
             def history():
                 for h in repos.get_path_history(path, rev):
                     if 'CHANGESET_VIEW' in req.perm(cset_resource(id=h[1])):
                         yield h
+
         elif revranges:
+            show_graph = path == '/' and not verbose \
+                         and not repos.has_linear_changesets \
+                         and len(revranges) == 1
+
             def history():
-                prevpath = path
-                expected_next_item = None
-                ranges = list(revranges.pairs)
-                ranges.reverse()
-                for (a, b) in ranges:
-                    a = repos.normalize_rev(a)
-                    b = repos.normalize_rev(b)
-                    while not repos.rev_older_than(b, a):
-                        node = get_existing_node(req, repos, prevpath, b)
-                        node_history = list(node.get_history(2))
-                        p, rev, chg = node_history[0]
+                separator = False
+                for a, b in reversed(revranges.pairs):
+                    curr_revrange[:] = (a, b)
+                    node = get_existing_node(req, repos, path, b)
+                    for p, rev, chg in node.get_history():
                         if repos.rev_older_than(rev, a):
-                            break # simply skip, no separator
+                            break
                         if 'CHANGESET_VIEW' in req.perm(cset_resource(id=rev)):
-                            if expected_next_item:
-                                # check whether we're continuing previous range
-                                np, nrev, nchg = expected_next_item
-                                if rev != nrev: # no, we need a separator
-                                    yield (np, nrev, None)
-                            yield node_history[0]
-                        prevpath = node_history[-1][0] # follow copy
-                        b = repos.previous_rev(rev)
-                        if len(node_history) > 1:
-                            expected_next_item = node_history[-1]
-                        else:
-                            expected_next_item = None
-                if expected_next_item:
-                    yield (expected_next_item[0], expected_next_item[1], None)
+                            separator = True
+                            yield p, rev, chg
+                    else:
+                        separator = False
+                    if separator:
+                        yield p, rev, None
         else:
+            show_graph = path == '/' and not verbose \
+                         and not repos.has_linear_changesets
+
             def history():
                 node = get_existing_node(req, repos, path, rev)
                 for h in node.get_history():
@@ -161,6 +164,7 @@ class LogModule(Component):
         depth = 1
         previous_path = normpath
         count = 0
+        history_remaining = True
         for old_path, old_rev, old_chg in history():
             if stop_rev and repos.rev_older_than(old_rev, stop_rev):
                 break
@@ -170,7 +174,7 @@ class LogModule(Component):
                 'path': old_path, 'rev': old_rev, 'existing_rev': old_rev,
                 'change': old_chg, 'depth': depth,
             }
-            
+
             if old_chg == Changeset.DELETE:
                 item['existing_rev'] = repos.previous_rev(old_rev, old_path)
             if not (mode == 'path_history' and old_chg == Changeset.EDIT):
@@ -184,7 +188,7 @@ class LogModule(Component):
                     break
                 elif mode == 'path_history':
                     depth -= 1
-            if old_chg is None: # separator entry
+            if old_chg is None:  # separator entry
                 stop_limit = limit
             else:
                 count += 1
@@ -192,13 +196,29 @@ class LogModule(Component):
             if count >= stop_limit:
                 break
             previous_path = old_path
-        if info == []:
+        else:
+            history_remaining = False
+        if not info:
             node = get_existing_node(req, repos, path, rev)
             if repos.rev_older_than(stop_rev, node.created_rev):
                 # FIXME: we should send a 404 error here
                 raise TracError(_("The file or directory '%(path)s' doesn't "
-                    "exist at revision %(rev)s or at any previous revision.", 
-                    path=path, rev=display_rev(rev)), _('Nonexistent path'))
+                                  "exist at revision %(rev)s or at any "
+                                  "previous revision.", path=path,
+                                  rev=repos.display_rev(rev)),
+                                _('Nonexistent path'))
+
+        # Generate graph data
+        graph = {}
+        if show_graph:
+            threads, vertices, columns = \
+                make_log_graph(repos, (item['rev'] for item in info))
+            graph.update(threads=threads, vertices=vertices, columns=columns,
+                         colors=self.graph_colors,
+                         line_width=0.04, dot_radius=0.1)
+            add_script(req, 'common/js/excanvas.js', ie_if='IE')
+            add_script(req, 'common/js/log_graph.js')
+            add_script_data(req, graph=graph)
 
         def make_log_href(path, **args):
             link_rev = rev
@@ -211,29 +231,34 @@ class LogModule(Component):
             return req.href.log(repos.reponame or None, path, **params)
 
         if format in ('rss', 'changelog'):
-            info = [i for i in info if i['change']] # drop separators
+            info = [i for i in info if i['change']]  # drop separators
             if info and count > limit:
                 del info[-1]
-        elif info and count >= limit:
+        elif info and history_remaining and count >= limit:
             # stop_limit reached, there _might_ be some more
             next_rev = info[-1]['rev']
             next_path = info[-1]['path']
             next_revranges = None
-            if revranges:
-                next_revranges = str(revranges.truncate(next_rev))
+            if curr_revrange:
+                new_revrange = (curr_revrange[0], next_rev) \
+                               if info[-1]['change'] else None
+                next_revranges = revranges.truncate(curr_revrange,
+                                                    new_revrange)
+                next_revranges = unicode(next_revranges) or None
             if next_revranges or not revranges:
-                older_revisions_href = make_log_href(next_path, rev=next_rev,
-                                                     revs=next_revranges)
+                older_revisions_href = make_log_href(
+                    next_path, rev=next_rev, revs=next_revranges)
                 add_link(req, 'next', older_revisions_href,
-                    _('Revision Log (restarting at %(path)s, rev. %(rev)s)',
-                    path=next_path, rev=display_rev(next_rev)))
+                         _('Revision Log (restarting at %(path)s, rev. '
+                           '%(rev)s)', path=next_path,
+                           rev=repos.display_rev(next_rev)))
             # only show fully 'limit' results, use `change == None` as a marker
             info[-1]['change'] = None
-        
+
         revisions = [i['rev'] for i in info]
         changes = get_changes(repos, revisions, self.log)
         extra_changes = {}
-        
+
         if format == 'changelog':
             for rev in revisions:
                 changeset = changes[rev]
@@ -244,38 +269,38 @@ class LogModule(Component):
                 files = []
                 actions = []
                 for cpath, kind, chg, bpath, brev in changeset.get_changes():
-                    files.append(chg == Changeset.DELETE and bpath or cpath)
+                    files.append(bpath if chg == Changeset.DELETE else cpath)
                     actions.append(chg)
                 cs['files'] = files
                 cs['actions'] = actions
                 extra_changes[rev] = cs
 
         data = {
-            'context': Context.from_request(req, 'source', path,
-                                            parent=repos.resource),
+            'context': web_context(req, 'source', path, parent=repos.resource),
             'reponame': repos.reponame or None, 'repos': repos,
             'path': path, 'rev': rev, 'stop_rev': stop_rev,
-            'display_rev': display_rev, 'revranges': revranges,
-            'mode': mode, 'verbose': verbose, 'limit' : limit,
+            'display_rev': repos.display_rev, 'revranges': revranges,
+            'mode': mode, 'verbose': verbose, 'limit': limit,
             'items': info, 'changes': changes, 'extra_changes': extra_changes,
-            'wiki_format_messages':
-            self.config['changeset'].getbool('wiki_format_messages')
+            'graph': graph,
+            'wiki_format_messages': self.config['changeset']
+                                    .getbool('wiki_format_messages')
         }
 
         if format == 'changelog':
             return 'revisionlog.txt', data, 'text/plain'
         elif format == 'rss':
             data['email_map'] = Chrome(self.env).get_email_map()
-            data['context'] = Context.from_request(req, 'source', 
-                                                   path, parent=repos.resource,
-                                                   absurls=True)
+            data['context'] = web_context(req, 'source',
+                                          path, parent=repos.resource,
+                                          absurls=True)
             return 'revisionlog.rss', data, 'application/rss+xml'
 
         item_ranges = []
         range = []
         for item in info:
-            if item['change'] is None: # separator
-                if range: # start new range
+            if item['change'] is None:  # separator
+                if range:  # start new range
                     range.append(item)
                     item_ranges.append(range)
                     range = []
@@ -296,17 +321,18 @@ class LogModule(Component):
 
         rss_href = make_log_href(path, format='rss', revs=revs,
                                  stop_rev=stop_rev)
-        add_link(req, 'alternate', rss_href, _('RSS Feed'),
+        add_link(req, 'alternate', auth_link(req, rss_href), _('RSS Feed'),
                  'application/rss+xml', 'rss')
         changelog_href = make_log_href(path, format='changelog', revs=revs,
                                        stop_rev=stop_rev)
-        add_link(req, 'alternate', changelog_href, _('ChangeLog'), 'text/plain')
+        add_link(req, 'alternate', changelog_href, _('ChangeLog'),
+                 'text/plain')
 
-        add_ctxtnav(req, _('View Latest Revision'), 
+        add_ctxtnav(req, _('View Latest Revision'),
                     href=req.href.browser(repos.reponame or None, path))
         if 'next' in req.chrome['links']:
             next = req.chrome['links']['next'][0]
-            add_ctxtnav(req, tag.span(tag.a(_('Older Revisions'), 
+            add_ctxtnav(req, tag.span(tag.a(_('Older Revisions'),
                                             href=next['href']),
                                       Markup(' &rarr;')))
 
@@ -314,9 +340,10 @@ class LogModule(Component):
 
     # IWikiSyntaxProvider methods
 
-    REV_RANGE = r"(?:%s|%s)" % (Ranges.RE_STR, ChangesetModule.CHANGESET_ID)
-    #                          int rev ranges or any kind of rev
-    
+    # int rev ranges or any kind of rev range
+    REV_RANGE = r"(?:%(int)s|%(cset)s(?:[:-]%(cset)s)?)" % \
+                {'int': Ranges.RE_STR, 'cset': ChangesetModule.CHANGESET_ID}
+
     def get_wiki_syntax(self):
         yield (
             # [...] form, starts with optional intertrac: [T... or [trac ...
@@ -332,6 +359,8 @@ class LogModule(Component):
     def get_link_resolvers(self):
         yield ('log', self._format_link)
 
+    LOG_LINK_RE = re.compile(r"([^@:]*)[@:]%s?" % REV_RANGE)
+
     def _format_link(self, formatter, ns, match, label, fullmatch=None):
         if ns == 'log1':
             groups = fullmatch.groupdict()
@@ -358,7 +387,7 @@ class LogModule(Component):
                 indexes = [sep in match and match.index(sep) for sep in ':@']
                 idx = min([i for i in indexes if i is not False])
                 path, revs = match[:idx], match[idx+1:]
-        
+
         rm = RepositoryManager(self.env)
         try:
             reponame, repos, path = rm.get_repository_by_path(path)
@@ -368,23 +397,19 @@ class LogModule(Component):
                     repos = rm.get_repository(reponame)
 
             if repos:
-                revranges = None
-                if any(c for c in ':-,' if c in revs):
-                    revranges = self._normalize_ranges(repos, path, revs)
-                    revs = None
-                if 'LOG_VIEW' in formatter.perm:
-                    if revranges:
-                        href = formatter.href.log(repos.reponame or None,
-                                                  path or '/', 
-                                                  revs=str(revranges))
+                if 'LOG_VIEW' in formatter.perm(repos.resource):
+                    reponame = repos.reponame or None
+                    path = path or '/'
+                    revranges = RevRanges(repos, revs)
+                    if revranges.has_ranges():
+                        href = formatter.href.log(reponame, path,
+                                                  revs=unicode(revranges))
                     else:
-                        try:
-                            rev = repos.normalize_rev(revs)
-                        except NoSuchChangeset:
-                            rev = None
-                        href = formatter.href.log(repos.reponame or None,
-                                                  path or '/', rev=rev)
-                    if query and (revranges or revs):
+                        # try to resolve if single rev
+                        repos.normalize_rev(revs)
+                        href = formatter.href.log(reponame, path,
+                                                  rev=revs or None)
+                    if query and '?' in href:
                         query = '&' + query[1:]
                     return tag.a(label, class_='source',
                                  href=href + query + fragment)
@@ -397,20 +422,114 @@ class LogModule(Component):
             errmsg = to_unicode(e)
         return tag.a(label, class_='missing source', title=errmsg)
 
-    LOG_LINK_RE = re.compile(r"([^@:]*)[@:]%s?" % REV_RANGE)
 
-    def _normalize_ranges(self, repos, path, revs):
-        ranges = revs.replace(':', '-')
-        try:
-            # fast path; only numbers
-            return Ranges(ranges, reorder=True) 
-        except ValueError:
-            # slow path, normalize each rev
-            splitted_ranges = re.split(r'([-,])', ranges)
+class RevRanges(object):
+
+    def __init__(self, repos, revs=None, resolve=False):
+        self.repos = repos
+        self.resolve = resolve
+        self.pairs = []
+        self.a = self.b = None
+        if revs:
+            self._append(revs)
+
+    def has_ranges(self):
+        n = len(self.pairs)
+        return n > 1 or n == 1 and self.a != self.b
+
+    def truncate(self, curr_pair, new_pair=None):
+        curr_pair = tuple(curr_pair)
+        if new_pair:
+            new_pair = tuple(new_pair)
+        revranges = RevRanges(self.repos, resolve=self.resolve)
+        pairs = revranges.pairs
+        for pair in self.pairs:
+            if pair == curr_pair:
+                if new_pair:
+                    pairs.append(new_pair)
+                break
+            pairs.append(pair)
+        if pairs:
+            revranges.a = pairs[0][0]
+            revranges.b = pairs[-1][1]
+        revranges._reduce()
+        return revranges
+
+    def _normrev(self, rev):
+        if not rev:
+            raise NoSuchChangeset(rev)
+        if self.resolve:
+            return self.repos.normalize_rev(rev)
+        elif self.repos.has_linear_changesets:
             try:
-                revs = [repos.normalize_rev(r) for r in splitted_ranges[::2]]
-            except NoSuchChangeset:
-                return None
-            seps = splitted_ranges[1::2] + ['']
-            ranges = ''.join([str(rev)+sep for rev, sep in zip(revs, seps)])
-            return Ranges(ranges)
+                return int(rev)
+            except (ValueError, TypeError):
+                return rev
+        else:
+            return rev
+
+    _cset_range_re = re.compile(r"""(?:
+        %(cset)s[:-]%(cset)s    |  # int or hexa revs
+        [0-9]+[:-][A-Za-z_0-9]+ |  # e.g. 42-head
+        [A-Za-z_0-9]+[:-][0-9]+ |  # e.g. head-42
+        [^:]+:[^:]+                # e.g. master:dev-42
+        )\Z
+        """ % {'cset': ChangesetModule.CHANGESET_ID}, re.VERBOSE)
+
+    def _append(self, revs):
+        if not revs:
+            return
+
+        pairs = []
+        for rev in re.split(u',\u200b?', revs):
+            a = b = None
+            if self._cset_range_re.match(rev):
+                for sep in ':-':
+                    if sep in rev:
+                        a, b = rev.split(sep)
+                        break
+            if a is None:
+                a = b = self._normrev(rev)
+            elif a == b:
+                a = b = self._normrev(a)
+            else:
+                a = self._normrev(a)
+                b = self._normrev(b)
+            pairs.append((a, b))
+        self.pairs.extend(pairs)
+        self._reduce()
+
+    def _reduce(self):
+        if all(isinstance(pair[0], (int, long)) and
+               isinstance(pair[1], (int, long))
+               for pair in self.pairs):
+            try:
+                ranges = Ranges(unicode(self), reorder=True)
+            except:
+                pass
+            else:
+                self.pairs[:] = ranges.pairs
+        else:
+            seen = set()
+            pairs = self.pairs[:]
+            for idx, pair in enumerate(pairs):
+                if pair in seen:
+                    pairs[idx] = None
+                else:
+                    seen.add(pair)
+            if len(pairs) != len(seen):
+                self.pairs[:] = filter(None, pairs)
+        if self.pairs:
+            self.a = self.pairs[0][0]
+            self.b = self.pairs[-1][1]
+        else:
+            self.a = self.b = None
+
+    def __len__(self):
+        return len(self.pairs)
+
+    def __unicode__(self):
+        sep = '-' if self.repos.has_linear_changesets else ':'
+        return ','.join(sep.join(map(unicode, pair)) if pair[0] != pair[1]
+                                                     else unicode(pair[0])
+                        for pair in self.pairs)
