@@ -16,13 +16,18 @@
 # Author: Jonas Borgström <jonas@edgewall.com>
 #         Christopher Lenz <cmlenz@gmx.de>
 
+import re
+
 from trac.util.text import unicode_quote, unicode_urlencode
 
 
+slashes_re = re.compile(r'/{2,}')
+
+
 class Href(object):
     """Implements a callable that constructs URLs with the given base. The
     function can be called with any number of positional and keyword
-    arguments which than are used to assemble the URL.
+    arguments which then are used to assemble the URL.
 
     Positional arguments are appended as individual segments to
     the path of the URL:
@@ -118,10 +123,25 @@ class Href(object):
     '/trac/ticket/540'
     >>> href.browser('/trunk/README.txt', format='txt')
     '/trac/browser/trunk/README.txt?format=txt'
+
+    The ``path_safe`` argument specifies the characters that don't
+    need to be quoted in the path arguments. Likewise, the
+    ``query_safe`` argument specifies the characters that don't need
+    to be quoted in the query string:
+
+    >>> href = Href('')
+    >>> href.milestone('<look,here>', param='<here,too>')
+    '/milestone/%3Clook%2Chere%3E?param=%3Chere%2Ctoo%3E'
+
+    >>> href = Href('', path_safe='/<,', query_safe=',>')
+    >>> href.milestone('<look,here>', param='<here,too>')
+    '/milestone/<look,here%3E?param=%3Chere,too>'
     """
 
-    def __init__(self, base):
+    def __init__(self, base, path_safe="/!~*'()", query_safe="!~*'()"):
         self.base = base.rstrip('/')
+        self.path_safe = path_safe
+        self.query_safe = query_safe
         self._derived = {}
 
     def __call__(self, *args, **kw):
@@ -147,19 +167,18 @@ class Href(object):
                 args = args[:-1]
 
         # build the path
-        path = '/'.join([unicode_quote(unicode(arg).strip('/')) for arg in args
-                         if arg is not None])
+        path = '/'.join(unicode_quote(unicode(arg).strip('/'), self.path_safe)
+                        for arg in args if arg is not None)
         if path:
-            href += '/' + path
+            href += '/' + slashes_re.sub('/', path).lstrip('/')
         elif not href:
             href = '/'
 
         # assemble the query string
         for k, v in kw.items():
-            add_param(k.endswith('_') and k[:-1] or k, v)
-
+            add_param(k[:-1] if k.endswith('_') else k, v)
         if params:
-            href += '?' + unicode_urlencode(params)
+            href += '?' + unicode_urlencode(params, self.query_safe)
 
         return href
 
@@ -168,12 +187,17 @@ class Href(object):
             self._derived[name] = lambda *args, **kw: self(name, *args, **kw)
         return self._derived[name]
 
+    _printable_safe = ''.join(map(chr, xrange(0x21, 0x7f)))
+
     def __add__(self, rhs):
-        if rhs.startswith('/'):
-            return self.base + rhs
-        if rhs:
-            return self.base + '/' + rhs
-        return self.base or '/'
+        if not rhs:
+            return self.base or '/'
+        if rhs.startswith('?'):
+            return (self.base or '/') + \
+                   unicode_quote(rhs, self._printable_safe)
+        if not rhs.startswith('/'):
+            rhs = '/' + rhs
+        return self.base + unicode_quote(rhs, self._printable_safe)
 
 
 if __name__ == '__main__':
