@@ -18,21 +18,25 @@
 # Author: Daniel Lundin <daniel@edgewall.com>
 #         Christopher Lenz <cmlenz@gmx.de>
 
-import time
+from __future__ import with_statement
 
-from trac.admin.api import console_date_format
+import re
+import sys
+
+from trac.admin.api import console_date_format, get_console_locale
 from trac.core import TracError, Component, implements
 from trac.util import hex_entropy
 from trac.util.text import print_table
 from trac.util.translation import _
-from trac.util.datefmt import format_date, parse_date, to_datetime, \
-                              to_timestamp
+from trac.util.datefmt import get_datetime_format_hint, format_date, \
+                              parse_date, time_now, to_datetime, to_timestamp
 from trac.admin.api import IAdminCommandProvider, AdminCommandError
 
 UPDATE_INTERVAL = 3600 * 24 # Update session last_visit time stamp after 1 day
 PURGE_AGE = 3600 * 24 * 90 # Purge session after 90 days idle
 COOKIE_KEY = 'trac_session'
 
+
 # Note: as we often manipulate both the `session` and the
 #       `session_attribute` tables, there's a possibility of table
 #       deadlocks (#9705). We try to prevent them to happen by always
@@ -55,34 +59,42 @@ class DetachedSession(dict):
     def __setitem__(self, key, value):
         dict.__setitem__(self, key, unicode(value))
 
+    def set(self, key, value, default=None):
+        """Set a variable in the session, or remove it if it's equal to the
+        default value.
+        """
+        value = unicode(value)
+        if default is not None:
+            default = unicode(default)
+            if value == default:
+                self.pop(key, None)
+                return
+        dict.__setitem__(self, key, value)
+
     def get_session(self, sid, authenticated=False):
-        self.env.log.debug('Retrieving session for ID %r', sid)
+        self.env.log.debug("Retrieving session for ID %r", sid)
 
-        db = self.env.get_db_cnx()
-        cursor = db.cursor()
+        with self.env.db_query as db:
+            self.sid = sid
+            self.authenticated = authenticated
+            self.clear()
 
-        self.sid = sid
-        self.authenticated = authenticated
-        self.clear()
-
-        cursor.execute("""
-            SELECT last_visit FROM session WHERE sid=%s AND authenticated=%s
-            """, (sid, int(authenticated)))
-        row = cursor.fetchone()
-        if not row:
-            self.last_visit = 0
-            self._new = True
-            self._old = {}
-            return
-        self._new = False
-        self.last_visit = int(row[0] or 0)
-
-        cursor.execute("""
-            SELECT name,value FROM session_attribute
-            WHERE sid=%s and authenticated=%s
-            """, (sid, int(authenticated)))
-        self.update(cursor)
-        self._old = self.copy()
+            for last_visit, in db("""
+                    SELECT last_visit FROM session
+                    WHERE sid=%s AND authenticated=%s
+                    """, (sid, int(authenticated))):
+                self._new = False
+                self.last_visit = int(last_visit or 0)
+                self.update(db("""
+                    SELECT name, value FROM session_attribute
+                    WHERE sid=%s and authenticated=%s
+                    """, (sid, int(authenticated))))
+                self._old = self.copy()
+                break
+            else:
+                self.last_visit = 0
+                self._new = True
+                self._old = {}
 
     def save(self):
         items = self.items()
@@ -92,20 +104,17 @@ class DetachedSession(dict):
             return
 
         authenticated = int(self.authenticated)
-        now = int(time.time())
+        now = int(time_now())
 
         # We can't do the session management in one big transaction,
         # as the intertwined changes to both the session and
         # session_attribute tables are prone to deadlocks (#9705).
         # Therefore we first we save the current session, then we
         # eventually purge the tables.
-        
-        session_saved = [False]
 
-        @self.env.with_transaction()
-        def save_session(db):
-            cursor = db.cursor()
+        session_saved = False
 
+        with self.env.db_transaction as db:
             # Try to save the session if it's a new one. A failure to
             # do so is not critical but we nevertheless skip the
             # following steps.
@@ -116,11 +125,10 @@ class DetachedSession(dict):
                 # The session might already exist even if _new is True since
                 # it could have been created by a concurrent request (#3563).
                 try:
-                    cursor.execute("""
-                        INSERT INTO session (sid,last_visit,authenticated)
-                        VALUES (%s,%s,%s)
-                        """, (self.sid, self.last_visit, authenticated))
-                except Exception:
+                    db("""INSERT INTO session (sid, last_visit, authenticated)
+                          VALUES (%s,%s,%s)
+                          """, (self.sid, self.last_visit, authenticated))
+                except self.env.db_exc.IntegrityError:
                     self.env.log.warning('Session %s already exists', self.sid)
                     db.rollback()
                     return
@@ -129,64 +137,59 @@ class DetachedSession(dict):
             # new ones. The last concurrent request to do so "wins".
 
             if self._old != self:
+                if self._old.get('name') != self.get('name') or \
+                        self._old.get('email') != self.get('email'):
+                    self.env.invalidate_known_users_cache()
                 if not items and not authenticated:
                     # No need to keep around empty unauthenticated sessions
-                    cursor.execute("""
-                        DELETE FROM session WHERE sid=%s AND authenticated=0
-                        """, (self.sid,))
-                cursor.execute("""
-                    DELETE FROM session_attribute
-                    WHERE sid=%s AND authenticated=%s
-                    """, (self.sid, authenticated))
+                    db("DELETE FROM session WHERE sid=%s AND authenticated=0",
+                       (self.sid,))
+                db("""DELETE FROM session_attribute
+                      WHERE sid=%s AND authenticated=%s
+                      """, (self.sid, authenticated))
                 self._old = dict(self.items())
                 # The session variables might already have been updated by a
                 # concurrent request.
                 try:
-                    cursor.executemany("""
+                    db.executemany("""
                         INSERT INTO session_attribute
                           (sid,authenticated,name,value)
                         VALUES (%s,%s,%s,%s)
                         """, [(self.sid, authenticated, k, v)
                               for k, v in items])
-                except Exception:
+                except self.env.db_exc.IntegrityError:
                     self.env.log.warning('Attributes for session %s already '
                                          'updated', self.sid)
                     db.rollback()
                     return
-                session_saved[0] = True
+                session_saved = True
 
         # Purge expired sessions. We do this only when the session was
         # changed as to minimize the purging.
 
-        if session_saved[0] and now - self.last_visit > UPDATE_INTERVAL:
+        if session_saved and now - self.last_visit > UPDATE_INTERVAL:
             self.last_visit = now
             mintime = now - PURGE_AGE
 
-            @self.env.with_transaction()
-            def purge_session_attributes(db):
-                cursor = db.cursor()
+            with self.env.db_transaction as db:
                 # Update the session last visit time if it is over an
                 # hour old, so that session doesn't get purged
                 self.env.log.info("Refreshing session %s", self.sid)
-                cursor.execute("""
-                    UPDATE session SET last_visit=%s
-                    WHERE sid=%s AND authenticated=%s
-                    """, (self.last_visit, self.sid, authenticated))
+                db("""UPDATE session SET last_visit=%s
+                      WHERE sid=%s AND authenticated=%s
+                      """, (self.last_visit, self.sid, authenticated))
                 self.env.log.debug('Purging old, expired, sessions.')
-                cursor.execute("""
-                    DELETE FROM session_attribute
-                    WHERE authenticated=0 AND sid IN (
-                      SELECT sid FROM session 
-                      WHERE authenticated=0 AND last_visit < %s
-                    )
-                    """, (mintime,))
+                db("""DELETE FROM session_attribute
+                      WHERE authenticated=0 AND sid IN (
+                          SELECT sid FROM session
+                          WHERE authenticated=0 AND last_visit < %s
+                      )
+                      """, (mintime,))
 
             # Avoid holding locks on lot of rows on both session_attribute
             # and session tables
-            @self.env.with_transaction()
-            def purge_sessions(db):
-                cursor = db.cursor()
-                cursor.execute("""
+            with self.env.db_transaction as db:
+                db("""
                     DELETE FROM session
                     WHERE authenticated=0 AND last_visit < %s
                     """, (mintime,))
@@ -199,14 +202,14 @@ class Session(DetachedSession):
         super(Session, self).__init__(env, None)
         self.req = req
         if req.authname == 'anonymous':
-            if not req.incookie.has_key(COOKIE_KEY):
+            if COOKIE_KEY not in req.incookie:
                 self.sid = hex_entropy(24)
                 self.bake_cookie()
             else:
                 sid = req.incookie[COOKIE_KEY].value
                 self.get_session(sid)
         else:
-            if req.incookie.has_key(COOKIE_KEY):
+            if COOKIE_KEY in req.incookie:
                 sid = req.incookie[COOKIE_KEY].value
                 self.promote_session(sid)
             self.get_session(req.authname, authenticated=True)
@@ -218,15 +221,21 @@ class Session(DetachedSession):
         self.req.outcookie[COOKIE_KEY]['expires'] = expires
         if self.env.secure_cookies:
             self.req.outcookie[COOKIE_KEY]['secure'] = True
+        if sys.version_info >= (2, 6):
+            self.req.outcookie[COOKIE_KEY]['httponly'] = True
+
+    _valid_sid_re = re.compile(r'[_A-Za-z0-9]+\Z')
 
     def get_session(self, sid, authenticated=False):
         refresh_cookie = False
 
+        if not authenticated and not self._valid_sid_re.match(sid):
+            raise TracError(_("Session ID must be alphanumeric."))
         if self.sid and sid != self.sid:
             refresh_cookie = True
 
         super(Session, self).get_session(sid, authenticated)
-        if self.last_visit and time.time() - self.last_visit > UPDATE_INTERVAL:
+        if self.last_visit and time_now() - self.last_visit > UPDATE_INTERVAL:
             refresh_cookie = True
 
         # Refresh the session cookie if this is the first visit after a day
@@ -239,24 +248,22 @@ class Session(DetachedSession):
         assert new_sid, 'Session ID cannot be empty'
         if new_sid == self.sid:
             return
-        @self.env.with_transaction()
-        def update_session_id(db):
-            cursor = db.cursor()
-            cursor.execute("SELECT sid FROM session WHERE sid=%s", (new_sid,))
-            if cursor.fetchone():
+        if not self._valid_sid_re.match(new_sid):
+            raise TracError(_("Session ID must be alphanumeric."),
+                            _("Error renaming session"))
+        with self.env.db_transaction as db:
+            if db("SELECT sid FROM session WHERE sid=%s", (new_sid,)):
                 raise TracError(_("Session '%(id)s' already exists. "
                                   "Please choose a different session ID.",
                                   id=new_sid),
                                 _("Error renaming session"))
-            self.env.log.debug('Changing session ID %s to %s', self.sid,
+            self.env.log.debug("Changing session ID %s to %s", self.sid,
                                new_sid)
-            cursor.execute("""
-                UPDATE session SET sid=%s WHERE sid=%s AND authenticated=0
+            db("UPDATE session SET sid=%s WHERE sid=%s AND authenticated=0",
+               (new_sid, self.sid))
+            db("""UPDATE session_attribute SET sid=%s
+                  WHERE sid=%s and authenticated=0
                   """, (new_sid, self.sid))
-            cursor.execute("""
-                UPDATE session_attribute SET sid=%s 
-                WHERE sid=%s and authenticated=0
-                """, (new_sid, self.sid))
         self.sid = new_sid
         self.bake_cookie()
 
@@ -265,51 +272,43 @@ class Session(DetachedSession):
         is no preexisting session data for that user name.
         """
         assert self.req.authname != 'anonymous', \
-               'Cannot promote session of anonymous user'
+               "Cannot promote session of anonymous user"
+
+        with self.env.db_transaction as db:
+            authenticated_flags = [authenticated for authenticated, in db(
+                "SELECT authenticated FROM session WHERE sid=%s OR sid=%s",
+                (sid, self.req.authname))]
 
-        @self.env.with_transaction()
-        def update_session_id(db):
-            cursor = db.cursor()
-            cursor.execute("""
-                SELECT authenticated FROM session WHERE sid=%s OR sid=%s
-                """, (sid, self.req.authname))
-            authenticated_flags = [row[0] for row in cursor.fetchall()]
-            
             if len(authenticated_flags) == 2:
                 # There's already an authenticated session for the user,
                 # we simply delete the anonymous session
-                cursor.execute("""
-                    DELETE FROM session WHERE sid=%s AND authenticated=0
-                    """, (sid,))
-                cursor.execute("""
-                    DELETE FROM session_attribute
-                    WHERE sid=%s AND authenticated=0
-                    """, (sid,))
+                db("DELETE FROM session WHERE sid=%s AND authenticated=0",
+                   (sid,))
+                db("""DELETE FROM session_attribute
+                      WHERE sid=%s AND authenticated=0
+                      """, (sid,))
             elif len(authenticated_flags) == 1:
                 if not authenticated_flags[0]:
                     # Update the anomymous session records so the session ID
                     # becomes the user name, and set the authenticated flag.
-                    self.env.log.debug('Promoting anonymous session %s to '
-                                       'authenticated session for user %s',
+                    self.env.log.debug("Promoting anonymous session %s to "
+                                       "authenticated session for user %s",
                                        sid, self.req.authname)
-                    cursor.execute("""
-                        UPDATE session SET sid=%s,authenticated=1
-                        WHERE sid=%s AND authenticated=0
-                        """, (self.req.authname, sid))
-                    cursor.execute("""
-                        UPDATE session_attribute SET sid=%s,authenticated=1
-                        WHERE sid=%s
-                        """, (self.req.authname, sid))
+                    db("""UPDATE session SET sid=%s, authenticated=1
+                          WHERE sid=%s AND authenticated=0
+                          """, (self.req.authname, sid))
+                    db("""UPDATE session_attribute SET sid=%s, authenticated=1
+                          WHERE sid=%s
+                          """, (self.req.authname, sid))
             else:
                 # We didn't have an anonymous session for this sid. The
                 # authenticated session might have been inserted between the
                 # SELECT above and here, so we catch the error.
                 try:
-                    cursor.execute("""
-                        INSERT INTO session (sid,last_visit,authenticated)
-                        VALUES (%s,%s,1)
-                        """, (self.req.authname, int(time.time())))
-                except Exception:
+                    db("""INSERT INTO session (sid, last_visit, authenticated)
+                          VALUES (%s, %s, 1)
+                          """, (self.req.authname, int(time_now())))
+                except self.env.db_exc.IntegrityError:
                     self.env.log.warning('Authenticated session for %s '
                                          'already exists', self.req.authname)
                     db.rollback()
@@ -325,6 +324,10 @@ class SessionAdmin(Component):
     implements(IAdminCommandProvider)
 
     def get_admin_commands(self):
+        hints = {
+           'datetime': get_datetime_format_hint(get_console_locale(self.env)),
+           'iso8601': get_datetime_format_hint('iso8601'),
+        }
         yield ('session list', '[sid[:0|1]] [...]',
                """List the name and email for the given sids
 
@@ -332,7 +335,7 @@ class SessionAdmin(Component):
                sessions, and 'authenticated' all authenticated sessions.
                '*' lists all sessions, and is the default if no sids are
                given.
-               
+
                An sid suffix ':0' operates on an unauthenticated session with
                the given sid, and a suffix ':1' on an authenticated session
                (the default).""",
@@ -349,7 +352,7 @@ class SessionAdmin(Component):
 
         yield ('session set', '<name|email> <sid[:0|1]> <value>',
                """Set the name or email attribute of the given sid
-               
+
                An sid suffix ':0' operates on an unauthenticated session with
                the given sid, and a suffix ':1' on an authenticated session
                (the default).""",
@@ -365,10 +368,11 @@ class SessionAdmin(Component):
                self._complete_delete, self._do_delete)
 
         yield ('session purge', '<age>',
-               """Purge all anonymous sessions older than the given age
+               """Purge anonymous sessions older than the given age or date
 
                Age may be specified as a relative time like "90 days ago", or
-               in YYYYMMDD format.""",
+               as a date in the "%(datetime)s" or "%(iso8601)s" (ISO 8601)
+               format.""" % hints,
                None, self._do_purge)
 
     def _split_sid(self, sid):
@@ -380,19 +384,15 @@ class SessionAdmin(Component):
             return (sid, 1)
 
     def _get_sids(self):
-        db = self.env.get_db_cnx()
-        cursor = db.cursor()
-        cursor.execute("SELECT sid, authenticated FROM session")
-        return ['%s:%d' % (sid, auth) for sid, auth in cursor]
+        rows = self.env.db_query("SELECT sid, authenticated FROM session")
+        return ['%s:%d' % (sid, auth) for sid, auth in rows]
 
     def _get_list(self, sids):
         all_anon = 'anonymous' in sids or '*' in sids
         all_auth = 'authenticated' in sids or '*' in sids
         sids = set(self._split_sid(sid) for sid in sids
                    if sid not in ('anonymous', 'authenticated', '*'))
-        db = self.env.get_db_cnx()
-        cursor = db.cursor()
-        cursor.execute("""
+        rows = self.env.db_query("""
             SELECT DISTINCT s.sid, s.authenticated, s.last_visit,
                             n.value, e.value
             FROM session AS s
@@ -404,7 +404,7 @@ class SessionAdmin(Component):
                     AND e.name='email')
             ORDER BY s.sid, s.authenticated
             """)
-        for sid, authenticated, last_visit, name, email in cursor:
+        for sid, authenticated, last_visit, name, email in rows:
             if all_anon and not authenticated or all_auth and authenticated \
                     or (sid, authenticated) in sids:
                 yield (sid, authenticated, last_visit, name, email)
@@ -432,83 +432,71 @@ class SessionAdmin(Component):
                      for r in self._get_list(sids)],
                     [_('SID'), _('Auth'), _('Last Visit'), _('Name'),
                      _('Email')])
-        
+
     def _do_add(self, sid, name=None, email=None):
         sid, authenticated = self._split_sid(sid)
-        @self.env.with_transaction()
-        def add_session(db):
-            cursor = db.cursor()
+        with self.env.db_transaction as db:
             try:
-                cursor.execute("INSERT INTO session VALUES (%s, %s, %s)",
-                               (sid, authenticated, int(time.time())))
+                db("INSERT INTO session VALUES (%s, %s, %s)",
+                   (sid, authenticated, int(time_now())))
             except Exception:
                 raise AdminCommandError(_("Session '%(sid)s' already exists",
                                           sid=sid))
             if name is not None:
-                cursor.execute("""
-                    INSERT INTO session_attribute VALUES (%s, %s, 'name', %s)
-                    """, (sid, authenticated, name))
+                db("INSERT INTO session_attribute VALUES (%s,%s,'name',%s)",
+                    (sid, authenticated, name))
             if email is not None:
-                cursor.execute("""
-                    INSERT INTO session_attribute VALUES (%s, %s, 'email', %s)
-                    """, (sid, authenticated, email))
+                db("INSERT INTO session_attribute VALUES (%s,%s,'email',%s)",
+                    (sid, authenticated, email))
+        self.env.invalidate_known_users_cache()
 
     def _do_set(self, attr, sid, val):
         if attr not in ('name', 'email'):
             raise AdminCommandError(_("Invalid attribute '%(attr)s'",
                                       attr=attr))
         sid, authenticated = self._split_sid(sid)
-        @self.env.with_transaction()
-        def set_attr(db):
-            cursor = db.cursor()
-            cursor.execute("""
-                SELECT sid FROM session WHERE sid=%s AND authenticated=%s
-                """, (sid, authenticated))
-            if not cursor.fetchone():
+        with self.env.db_transaction as db:
+            if not db("""SELECT sid FROM session
+                         WHERE sid=%s AND authenticated=%s""",
+                         (sid, authenticated)):
                 raise AdminCommandError(_("Session '%(sid)s' not found",
                                           sid=sid))
-            cursor.execute("""
+            db("""
                 DELETE FROM session_attribute
                 WHERE sid=%s AND authenticated=%s AND name=%s
                 """, (sid, authenticated, attr))
-            cursor.execute("""
-                INSERT INTO session_attribute VALUES (%s, %s, %s, %s)
-                """, (sid, authenticated, attr, val))
+            db("INSERT INTO session_attribute VALUES (%s, %s, %s, %s)",
+               (sid, authenticated, attr, val))
+        self.env.invalidate_known_users_cache()
 
     def _do_delete(self, *sids):
-        @self.env.with_transaction()
-        def delete_session(db):
-            cursor = db.cursor()
+        with self.env.db_transaction as db:
             for sid in sids:
                 sid, authenticated = self._split_sid(sid)
                 if sid == 'anonymous':
-                    cursor.execute("""
-                        DELETE FROM session WHERE authenticated=0
-                        """)
-                    cursor.execute("""
-                        DELETE FROM session_attribute WHERE authenticated=0
-                        """)
+                    db("DELETE FROM session WHERE authenticated=0")
+                    db("DELETE FROM session_attribute WHERE authenticated=0")
                 else:
-                    cursor.execute("""
+                    db("""
                         DELETE FROM session
                         WHERE sid=%s AND authenticated=%s
                         """, (sid, authenticated))
-                    cursor.execute("""
+                    db("""
                         DELETE FROM session_attribute
                         WHERE sid=%s AND authenticated=%s
                         """, (sid, authenticated))
+        self.env.invalidate_known_users_cache()
 
     def _do_purge(self, age):
-        when = parse_date(age)
-        @self.env.with_transaction()
-        def purge_session(db):
-            cursor = db.cursor()
+        when = parse_date(age, hint='datetime',
+                          locale=get_console_locale(self.env))
+        with self.env.db_transaction as db:
             ts = to_timestamp(when)
-            cursor.execute("""
+            db("""
                 DELETE FROM session
                 WHERE authenticated=0 AND last_visit<%s
                 """, (ts,))
-            cursor.execute("""
+            db("""
                 DELETE FROM session_attribute
                 WHERE authenticated=0
                       AND sid NOT IN (SELECT sid FROM session
